# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import atexit
import os
import tempfile
from datetime import datetime
from typing import Optional, Sequence, Union

import cv2
import numpy as np

from geti_sdk.data_models import Dataset, MediaType, Video
from geti_sdk.data_models.containers import MediaList
from geti_sdk.http_session import GetiRequestException
from geti_sdk.rest_converters import MediaRESTConverter

from .media_client import BaseMediaClient


class VideoClient(BaseMediaClient[Video]):
    """
    Class to manage video uploads and downloads for a certain project
    """

    _MEDIA_TYPE = MediaType.VIDEO

    def get_all_videos(self, dataset: Optional[Dataset] = None) -> MediaList[Video]:
        """
        Get the ID's and filenames of all videos in the project, from a specific
        dataset. If no dataset is passed, videos from the training dataset will be
        returned

        :param dataset: Dataset for which to retrieve the videos. If no dataset is
            passed, videos from the training dataset are returned.
        :return: A list containing all Video's in the project
        """
        return self._get_all(dataset=dataset)

    def upload_video(
        self,
        video: Union[np.ndarray, str, os.PathLike],
        dataset: Optional[Dataset] = None,
    ) -> Video:
        """
        Upload a video file to the server. Accepts either a path to a video file, or
        a numpy array containing pixel data for video frames.

        In case a numpy array is passed, this method expects the array to be 4
        dimensional, it's dimensions shaped as: [frames, heigth, width, channels]. The
        framerate of the created video will be set to 1 fps.

        :param video: full path to the video on disk, or numpy array holding the video
            pixel data
        :param dataset: Dataset to which to upload the video. If no dataset is
            passed, the video is uploaded to the training dataset
        :return: Video object representing the uploaded video on the server
        """
        temporary_file_created = False
        if isinstance(video, (str, os.PathLike)):
            video_path = video
        elif isinstance(video, np.ndarray):
            try:
                n_frames, frame_height, frame_width, channels = video.shape
            except GetiRequestException as error:
                raise ValueError(
                    f"Invalid video input shape, expected a 4D numpy array with "
                    f"dimensions representing [frames, height, width, channels]. Got "
                    f"shape {video.shape}"
                ) from error
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
            video_file = tempfile.NamedTemporaryFile(
                prefix="geti-sdk_temp_video_", suffix=f"_{timestamp}.avi", delete=False
            )
            # Close the file, opencv will open it again from the path
            video_file.close()
            video_path = video_file.name
            out = cv2.VideoWriter(
                video_path,
                cv2.VideoWriter_fourcc("M", "J", "P", "G"),
                1,
                (frame_width, frame_height),
            )
            for frame in video[:, ...]:
                out.write(frame)
            out.release()
            temporary_file_created = True
        else:
            raise TypeError(f"Invalid video type: {type(video)}.")

        video_dict = self._upload(video_path, dataset=dataset)
        uploaded_video = MediaRESTConverter.from_dict(
            input_dict=video_dict, media_type=Video
        )
        uploaded_video._data = video_path
        if temporary_file_created:
            uploaded_video._needs_tempfile_deletion = True

            # Register cleanup function on system exit to ensure __del__ gets called
            def clean_temp_video():
                uploaded_video.__del__()

            atexit.register(clean_temp_video)
        return uploaded_video

    def upload_folder(
        self,
        path_to_folder: str,
        n_videos: int = -1,
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
        max_threads: int = 5,
    ) -> MediaList[Video]:
        """
        Upload all videos in a folder to the project. Returns the mapping of video
        filename to the unique ID assigned by Intel Geti.

        :param path_to_folder: Folder with videos to upload
        :param n_videos: Number of videos to upload from folder
        :param skip_if_filename_exists: Set to True to skip uploading of a video
            if a video with the same filename already exists in the project.
            Defaults to False
        :param dataset: Dataset to which to upload the video. If no dataset is
            passed, the video is uploaded to the training dataset
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: MediaList containing all video's in the project
        """
        return self._upload_folder(
            path_to_folder=path_to_folder,
            n_media=n_videos,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
            max_threads=max_threads,
        )

    def download_all(
        self,
        path_to_folder: str,
        append_video_uid: bool = False,
        max_threads: int = 10,
        dataset: Optional[Dataset] = None,
    ) -> None:
        """
        Download all videos in a project to a folder on the local disk.

        :param path_to_folder: path to the folder in which the videos should be saved
        :param append_video_uid: True to append the UID of a video to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{video_id}'). If there are videos in the project with
            duplicate filename, this must be set to True to ensure all videos are
            downloaded. Otherwise videos with the same name will be skipped.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :param dataset: Dataset from which to download the videos. If no dataset is
            passed, videos from all datasets are downloaded.
        """
        self._download_all(
            path_to_folder,
            append_media_uid=append_video_uid,
            max_threads=max_threads,
            dataset=dataset,
        )

    def delete_videos(self, videos: Sequence[Video]) -> bool:
        """
        Delete all Video entities in `videos` from the project.

        :param videos: List of Video entities to delete
        :return: True if all videos on the list were deleted successfully,
            False otherwise
        """
        return self._delete_media(media_list=videos)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from collections import UserList
from typing import Any, Dict, List, Optional, Sequence

from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.project import Project
from geti_sdk.utils.serialization_helpers import deserialize_dictionary


class LabelList(UserList):
    """
    A list containing labels for an Intel® Geti™ inference model.
    """

    def __init__(self, data: Optional[Sequence[Label]] = None):
        self.data: List[Label] = []
        if data is not None:
            super().__init__(list(data))

        self._id_mapping: Dict[str, Label] = {}
        self._name_mapping: Dict[str, Label] = {}
        self._generate_indices()
        self._empty_label = next((x for x in self.data if x.is_empty), None)

    @property
    def has_empty_label(self) -> bool:
        """
        Return True if the list of Labels contains an empty label
        """
        return self._empty_label is not None

    def _generate_indices(self):
        """
        Map names and ID's to Label objects to enable quick label retrieval
        """
        self._id_mapping = {x.id: x for x in self.data}
        self._name_mapping = {x.name: x for x in self.data}

    def get_by_id(self, id: str) -> Label:
        """
        Return the Label object with ID corresponding to `id`
        """
        label = self._id_mapping.get(id, None)
        if label is None:
            raise KeyError(f"Label with id `{id}` was not found in the LabelList")
        return label

    def get_by_name(self, name: str) -> Label:
        """
        Return the Label object named `name`
        """
        label = self._name_mapping.get(name, None)
        if label is None:
            raise KeyError(f"Label named `{name}` was not found in the LabelList")
        return label

    @classmethod
    def from_json(cls, input_json: List[Dict[str, Any]]) -> "LabelList":
        """
        Create a LabelList object from json input. Input should be formatted as a list
        of dictionaries, each representing a single Label
        """
        label_list: List[Label] = []
        for item in input_json:
            label_list.append(deserialize_dictionary(item, Label))
        return cls(label_list)

    def create_scored_label(self, id_or_name: str, score: float) -> ScoredLabel:
        """
        Return a ScoredLabel object corresponding to the label identified by
        `id_or_name`, and with an assigned probability score of `score`

        :param id_or_name: ID or name of the Label to assign
        :param score: probability score of the label
        """
        try:
            label = self.get_by_id(id_or_name)
        except KeyError:
            label = self.get_by_name(id_or_name)
        return ScoredLabel.from_label(label, probability=score)

    def get_empty_label(self) -> Optional[Label]:
        """
        Return the empty label, if the LabelList contains one. If not, return None
        """
        return self._empty_label

    def get_non_empty_labels(self) -> "LabelList":
        """
        Return all non-empty labels
        """
        return LabelList([x for x in self.data if not x.is_empty])

    @classmethod
    def from_project(cls, project: Project, task_index: int = 0) -> "LabelList":
        """
        Create a LabelList object for the 'project', corresponding to the trainable
        task addressed by `task_index`

        :param project: Project for which to get the list of labels
        :param task_index: Index of the task for which to get the list of labels.
            Defaults to 0, i.e. the first trainable task in the project
        """
        task = project.pipeline.trainable_tasks[task_index]
        return cls(task.labels)

    def sort_by_ids(self, label_ids: List[str]):
        """
        Sort the labels in the LabelList by their ID, according to the order defined
        in `label_ids`
        """
        new_data: List[Label] = []
        for label_id in label_ids:
            if label_id is None:
                # Certain models have the label id as 'None' to signal an empty label
                if self.get_empty_label() is not None:
                    new_data.append(self.get_empty_label())
                else:
                    continue
            new_data.append(self.get_by_id(label_id))
        self.data = new_data



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os

BASE_TEST_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DUMMY_HOST = "dummy_host"
CASSETTE_BASE_PATH = os.path.join(BASE_TEST_PATH, "fixtures", "cassettes")
CASSETTE_PATH_KEY = "VERSIONED_CASSETTES_PATH"
RECORD_CASSETTE_KEY = "RECORD_CASSETTES_TEMPORARY_DIR"
CASSETTE_EXTENSION = "cassette"
PROJECT_PREFIX = "geti_sdk_test"
DUMMY_USER = "dummy_user"
DUMMY_PASSWORD = "dummy_password"  # nosec B105
DUMMY_TOKEN = "dummy_token"  # nosec B105
DUMMY_WORKSPACE_ID = "dummy_workspace_id"
DUMMY_ORGANIZATION_ID = "dummy_organization_id"



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from typing import Any, Dict, List, Optional, Union

from geti_sdk.data_models import (
    Algorithm,
    Dataset,
    Job,
    Project,
    ProjectStatus,
    Task,
    TaskConfiguration,
)
from geti_sdk.data_models.containers import AlgorithmList
from geti_sdk.data_models.enums import JobState, JobType
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import (
    ConfigurationRESTConverter,
    JobRESTConverter,
    StatusRESTConverter,
)
from geti_sdk.utils import get_supported_algorithms
from geti_sdk.utils.job_helpers import get_job_with_timeout, monitor_job, monitor_jobs


class TrainingClient:
    """
    Class to manage training jobs for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project.id}"
        self.supported_algos = get_supported_algorithms(
            rest_session=session, project=project, workspace_id=workspace_id
        )

    def get_status(self) -> ProjectStatus:
        """
        Get the current status of the project from the Intel® Geti™ server.

        :return: ProjectStatus object reflecting the current project status
        """
        response = self.session.get_rest_response(
            url=f"{self.base_url}/status", method="GET"
        )
        return StatusRESTConverter.from_dict(response)

    def is_training(self) -> bool:
        """
        Request the project status and return True if the project is training
        """
        return self.get_status().is_training

    def get_jobs(
        self, project_only: bool = True, running_only: bool = False
    ) -> List[Job]:
        """
        Return a list of all jobs on the Intel® Geti™ server.

        If `project_only = True` (the default), only those jobs related to the project
        managed by this TrainingClient will be returned. If set to False, all jobs in
        the workspace are returned.

        :param project_only: True to return only those jobs pertaining to the project
            for which the TrainingClient is active. False to return all jobs in the
            Intel® Geti™ workspace.
        :param running_only: If set to True, only return those jobs that are still
            running. Completed or Scheduled jobs will not be included in that case
        :return: List of Jobs
        """
        query = "?limit=100"
        if project_only:
            query += f"&project_id={self.project.id}"
        if running_only:
            query += "&state=running"
        response_list_key = "jobs"

        job_rest_list: List[dict] = []
        while response := self.session.get_rest_response(
            url=f"workspaces/{self.workspace_id}/jobs{query}&skip={len(job_rest_list)}",
            method="GET",
        ):
            job_rest_list.extend(response[response_list_key])
            job_count_dict = response["jobs_count"]
            total_job_count = (
                job_count_dict.get("n_running_jobs", 0)
                + job_count_dict.get("n_finished_jobs", 0)
                + job_count_dict.get("n_failed_jobs", 0)
                + job_count_dict.get("n_cancelled_jobs", 0)
            )
            if len(job_rest_list) >= total_job_count:
                break

        job_list: List[Job] = []
        for job_dict in job_rest_list:
            job = JobRESTConverter.from_dict(job_dict)
            if running_only and not job.is_running:
                continue
            job.workspace_id = self.workspace_id
            job.geti_version = self.session.version
            job_list.append(job)

        return job_list

    def get_algorithms_for_task(self, task: Union[Task, int]) -> AlgorithmList:
        """
        Return a list of supported algorithms for a specific task.

        The `task` parameter accepts both a Task object and an integer. If an int is
        passed, this will be considered the index of the task in the list of trainable
        tasks for the project which is managed by the TrainingClient.

        :param task: Task to get the supported algorithms for. If an integer is passed,
            this is considered the index of the task in the trainable task list of the
            project. So passing `task=0` will return the algorithms for the first
            trainable task, etc.
        :return: List of supported algorithms for the task
        """
        if isinstance(task, int):
            task = self.project.get_trainable_tasks()[task]
        return self.supported_algos.get_by_task_type(task.type)

    def train_task(
        self,
        task: Union[Task, int],
        dataset: Optional[Dataset] = None,
        algorithm: Optional[Algorithm] = None,
        train_from_scratch: bool = False,
        hyper_parameters: Optional[TaskConfiguration] = None,
        hpo_parameters: Optional[Dict[str, Any]] = None,
        await_running_jobs: bool = True,
        timeout: int = 3600,
    ) -> Job:
        """
        Start training of a specific task in the project.

        The `task` parameter accepts both a Task object and an integer. If an int is
        passed, this will be considered the index of the task in the list of trainable
        tasks for the project which is managed by the TrainingClient.

        :param task: Task or index of Task to train
        :param dataset: Optional Dataset to train on
        :param algorithm: Optional Algorithm to use in training. If left as None (the
            default), the default algorithm for the task will be used.
        :param train_from_scratch: True to train the model from scratch, False to
            continue training from an existing checkpoint (if any)
        :param hyper_parameters: Optional hyper parameters to use for training
        :param hpo_parameters: Optional set of parameters to use for automatic hyper
            parameter optimization. Only supported for version 1.1 and up
        :param await_running_jobs: True to wait for currently running jobs to
            complete. This will guarantee that the training request can be submitted
            successfully. Setting this to False will cause an error to be raised when
            a training request is submitted for a task for which a training job is
            already in progress.
        :param timeout: Timeout (in seconds) to wait for the Job to be created. If a
            training request is submitted successfully, a training job should be
            instantiated on the Geti server. If the Job does not appear on the server
            job list within the `timeout`, an error will be raised. This parameter only
            takes effect when `await_running_jobs` is set to True.
        :return: The training job that has been created
        """
        if isinstance(task, int):
            task = self.project.get_trainable_tasks()[task]
        if dataset is not None:
            logging.warning(
                "Training on a Dataset other than the default training dataset is not "
                "supported in the version of the Geti platform running on your server. "
                "The `dataset` parameter will be disregarded."
            )
        if dataset is None:
            dataset = self.project.training_dataset
        if algorithm is None:
            algorithm = self.supported_algos.get_default_for_task_type(task.type)
        request_data: Dict[str, Any] = {
            "dataset_id": dataset.id,
            "task_id": task.id,
            "train_from_scratch": train_from_scratch,
            "model_template_id": algorithm.model_template_id,
        }
        if hyper_parameters is not None:
            hypers = hyper_parameters.model_configurations
            hypers_rest = (
                ConfigurationRESTConverter.configurable_parameter_list_to_rest(hypers)
            )
            request_data.update({"hyper_parameters": hypers_rest})
        if hpo_parameters is not None:
            request_data.update(
                {
                    "enable_hyper_parameter_optimization": True,
                    "hpo_parameters": hpo_parameters,
                }
            )

        request_data.pop("dataset_id")
        data = request_data

        task_jobs = self.get_jobs_for_task(task=task)
        task_training_jobs = [job for job in task_jobs if job.type == JobType.TRAIN]
        log_warning_msg = ""
        if len(task_training_jobs) >= 1:
            if len(task_training_jobs) == 1:
                msg_start = f"A training job for task '{task.title}' is"
            else:
                msg_start = f"Multiple training jobs for task '{task.title}' are"
            log_warning_msg = msg_start + " already in progress on the server."

        if len(task_training_jobs) >= 1:
            if not await_running_jobs:
                raise RuntimeError(
                    log_warning_msg + " Unable to submit training request. Please "
                    "wait for the current training job to finish or "
                    "set the `await_running_jobs` parameter in this "
                    "method to `True`."
                )
            else:
                logging.info(
                    log_warning_msg + f" Awaiting completion of currently running jobs "
                    f"before a new train request can be submitted. "
                    f"Maximum waiting time set to {timeout} seconds."
                )
                self.monitor_jobs(jobs=task_training_jobs, timeout=timeout)

        response = self.session.get_rest_response(
            url=f"{self.base_url}/train", method="POST", data=data
        )

        job_id = response["job_id"]

        try:
            job = get_job_with_timeout(
                job_id=job_id,
                session=self.session,
                workspace_id=self.workspace_id,
                job_type="training",
            )
        except RuntimeError:
            raise RuntimeError(
                "Training job was submitted, but the TrainingClient was unable to "
                "find the resulting job on the platform."
            )
        logging.info(f"Training job with id {job.id} submitted successfully.")
        return job

    def monitor_jobs(
        self, jobs: List[Job], timeout: int = 10000, interval: int = 15
    ) -> List[Job]:
        """
        Monitor and print the progress of all jobs in the list `jobs`. Execution is
        halted until all jobs have either finished, failed or were cancelled.

        Progress will be reported in 15s intervals

        :param jobs: List of jobs to monitor
        :param timeout: Timeout (in seconds) after which to stop the monitoring
        :param interval: Time interval (in seconds) at which the TrainingClient polls
            the server to update the status of the jobs. Defaults to 15 seconds
        :return: List of finished (or failed) jobs with their status updated
        """
        return monitor_jobs(
            session=self.session, jobs=jobs, timeout=timeout, interval=interval
        )

    def monitor_job(self, job: Job, timeout: int = 10000, interval: int = 15) -> Job:
        """
        Monitor and print the progress of a `job`. Program execution is
        halted until the job has either finished, failed or was cancelled.

        Progress will be reported in 15s intervals

        :param job: job to monitor
        :param timeout: Timeout (in seconds) after which to stop the monitoring
        :param interval: Time interval (in seconds) at which the TrainingClient polls
            the server to update the status of the jobs. Defaults to 15 seconds
        :return: job with it's status updated
        """
        return monitor_job(
            session=self.session, job=job, timeout=timeout, interval=interval
        )

    def get_jobs_for_task(self, task: Task, running_only: bool = True) -> List[Job]:
        """
        Return a list of current jobs for the task, if any

        :param task: Task to retrieve the jobs for
        :param running_only: True to return only jobs that are currently running,
            False to return all jobs (including cancelled, finished or errored jobs)
        :return: List of Jobs running on the server for this particular task
        """
        project_jobs = self.get_jobs(project_only=True)
        task_jobs: List[Job] = []
        for job in project_jobs:
            if job.metadata is not None:
                if job.metadata.task is not None:
                    if job.metadata.task.task_id == task.id:
                        if running_only:
                            if job.state == JobState.RUNNING:
                                task_jobs.append(job)
                        else:
                            task_jobs.append(job)
        return task_jobs



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `import-export` package contains the `GetiIE` class with number of methods for importing and
exporting projets and datasets to and from the Intel® Geti™ platform.

Module contents
---------------
"""

from .import_export_module import GetiIE
from .tus_uploader import TUSUploader

__all__ = [
    "TUSUploader",
    "GetiIE",
]



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.
import logging
from copy import copy
from typing import Dict, List, Optional, Tuple, cast

import cv2
import numpy as np

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.label import ScoredLabel
from geti_sdk.data_models.shapes import Point, Polygon

logger = logging.getLogger(__name__)

Contour = List[Tuple[float, float]]
ContourInternal = Optional[List[Tuple[float, float]]]


def get_subcontours(contour: Contour) -> List[Contour]:
    """
    Split contour into sub-contours that do not have self intersections.

    :param contour: the contour to split
    :return: list of sub-contours
    """

    def find_loops(points: ContourInternal) -> List:
        """For each consecutive pair of equivalent rows in the input matrix returns their indices."""
        _, inverse, count = np.unique(points, axis=0, return_inverse=True, return_counts=True)  # type: ignore
        duplicates = np.where(count > 1)[0]
        indices = []
        for x in duplicates:
            y = np.nonzero(inverse == x)[0]
            for i, _ in enumerate(y[:-1]):
                indices.append(y[i : i + 2])
        return indices

    base_contour = cast(ContourInternal, copy(contour))

    # Make sure that contour is closed.
    if not np.array_equal(base_contour[0], base_contour[-1]):  # type: ignore
        base_contour.append(base_contour[0])

    subcontours: List[Contour] = []
    loops = sorted(find_loops(base_contour), key=lambda x: x[0], reverse=True)
    for loop in loops:
        i, j = loop
        subcontour = base_contour[i:j]
        subcontour = [x for x in subcontour if x is not None]
        subcontours.append(cast(Contour, subcontour))
        base_contour[i:j] = [None] * (j - i)

    return [i for i in subcontours if len(i) > 2]


def create_annotation_from_segmentation_map(
    hard_prediction: np.ndarray, soft_prediction: np.ndarray, label_map: Dict
) -> List[Annotation]:
    """
    Create polygons from the soft predictions.

    Note: background label will be ignored and not be converted to polygons.

    :param hard_prediction: hard prediction containing the final label index per pixel.
        See function `create_hard_prediction_from_soft_prediction`.
    :param soft_prediction: soft prediction with shape H x W x N_labels,
            where soft_prediction[:, :, 0] is the soft prediction for
            background. If soft_prediction is of H x W shape, it is
            assumed that this soft prediction will be applied for all
            labels.
    :param label_map: dictionary mapping labels to an index. It is assumed
            that the first item in the dictionary corresponds to the
            background label and will therefore be ignored.
    :return: list of annotations with polygons
    """
    # pylint: disable=too-many-locals
    img_class = hard_prediction.swapaxes(0, 1)

    # pylint: disable=too-many-nested-blocks
    annotations: List[Annotation] = []
    for label_index, label in label_map.items():
        # Skip background
        if label_index == 0:
            continue

        # obtain current label soft prediction
        if len(soft_prediction.shape) == 3:
            current_label_soft_prediction = soft_prediction[:, :, label_index]
        else:
            current_label_soft_prediction = soft_prediction

        obj_group = img_class == label_index
        label_index_map = (obj_group.T.astype(int) * 255).astype(np.uint8)

        # Contour retrieval mode CCOMP (Connected components) creates a two-level
        # hierarchy of contours
        contours, hierarchies = cv2.findContours(
            label_index_map, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE
        )

        if hierarchies is not None:
            for contour, hierarchy in zip(contours, hierarchies[0]):
                if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                    continue

                if hierarchy[3] == -1:
                    # In this case a contour does not represent a hole
                    _contour = [(point[0][0], point[0][1]) for point in contour]

                    # Split contour into subcontours that do not have self intersections.
                    subcontours = get_subcontours(_contour)
                    for subcontour in subcontours:
                        # compute probability of the shape
                        mask = np.zeros(hard_prediction.shape, dtype=np.uint8)
                        cv2.drawContours(
                            mask,
                            np.asarray([[[x, y]] for x, y in subcontour]),
                            contourIdx=-1,
                            color=1,
                            thickness=-1,
                        )
                        probability = cv2.mean(current_label_soft_prediction, mask)[0]

                        # convert the list of points to a closed polygon
                        points = [Point(x=x, y=y) for x, y in subcontour]
                        polygon = Polygon(points=points)

                        if polygon.area > 0:
                            # Contour is a closed polygon with area > 0
                            annotations.append(
                                Annotation(
                                    shape=polygon,
                                    labels=[ScoredLabel.from_label(label, probability)],
                                    # id=ID(ObjectId()),
                                )
                            )
                        else:
                            # Contour is a closed polygon with area == 0
                            logger.warning(
                                "The geometry of the segmentation map you are converting "
                                "is not fully supported. Polygons with a area of zero "
                                "will be removed.",
                            )
                else:
                    # If contour hierarchy[3] != -1 then contour has a parent and
                    # therefore is a hole
                    # Do not allow holes in segmentation masks to be filled silently,
                    # but trigger warning instead
                    logger.warning(
                        "The geometry of the segmentation map you are converting is "
                        "not fully supported. A hole was found and will be filled.",
                    )
    return annotations



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

BASE_TEMPLATE = {
    "name": "dummy project name",
    "pipeline": {
        "connections": [],
        "tasks": [{"title": "Dataset", "task_type": "dataset"}],
    },
}

DETECTION_TASK = {"title": "Detection task", "task_type": "detection", "labels": []}

ROTATED_DETECTION_TASK = {
    "title": "Rotated detection task",
    "task_type": "rotated_detection",
    "labels": [],
}

SEGMENTATION_TASK = {
    "title": "Segmentation task",
    "task_type": "segmentation",
    "labels": [],
}

INSTANCE_SEGMENTATION_TASK = {
    "title": "Instance segmentation task",
    "task_type": "instance_segmentation",
    "labels": [],
}

CLASSIFICATION_TASK = {
    "title": "Classification task",
    "task_type": "classification",
    "labels": [],
}

ANOMALY_CLASSIFICATION_TASK = {
    "title": "Anomaly classification task",
    "task_type": "anomaly_classification",
    "labels": [],
}

# This is the reduced anomaly task.
# It goes under `Anomaly` title,
# and it is `Anomally classification task` under the hood
ANOMALY_TASK = {
    "title": "Anomaly",
    "task_type": "anomaly",
    "labels": [],
}

ANOMALY_DETECTION_TASK = {
    "title": "Anomaly detection task",
    "task_type": "anomaly_detection",
    "labels": [],
}

ANOMALY_SEGMENTATION_TASK = {
    "title": "Anomaly segmentation task",
    "task_type": "anomaly_segmentation",
    "labels": [],
}

INSTANCE_SEGMENTATION_TASK = {
    "title": "Instance segmentation task",
    "task_type": "instance_segmentation",
    "labels": [],
}

ROTATED_DETECTION_TASK = {
    "title": "Rotated detection task",
    "task_type": "rotated_detection",
    "labels": [],
}

CROP_TASK = {"title": "Crop task", "task_type": "crop"}



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Dict, Optional

import attrs

DEFAULT_API_VERSION = "v1"
LEGACY_API_VERSION = "v1.0"


def trim_trailing_slash(input_string: str) -> str:
    """
    Remove trailing slash(es) from a string, if present.

    :param input_string: String to remove the trailing slash(es) from
    :return: Updated string
    """
    return input_string.rstrip("/")


@attrs.define(slots=False)
class ServerConfig:
    """
    Base configuration holding the connection details of the Intel® Geti™ server.
    Contains the hostname, ssl certificate configuration and proxy configuration.

    :var host: full hostname or ip address of the Intel® Geti™ server.
        Note: this should include the protocol (i.e. https://your_geti_hostname.com)
    :var has_valid_certificate: Set to True if the server has a valid SSL certificate
        that should be validated and used to establish an encrypted HTTPS connection
    :var proxies: Optional dictionary containing proxy information, if this is
        required to connect to the server. For example:

        proxies = {
            'http': http://proxy-server.com:<http_port_number>,
            'https': http://proxy-server.com:<https_port_number>
        },

        if set to None (the default), the global proxy settings found on the system
        will be used. If set to an emtpy dictionary, no proxy will be used.
    """

    host: str = attrs.field(converter=trim_trailing_slash)
    has_valid_certificate: bool = attrs.field(default=False, kw_only=True)
    proxies: Optional[Dict[str, str]] = attrs.field(default=None, kw_only=True)

    def __attrs_post_init__(self):
        """
        Initialize private attributes
        """
        self._api_version = DEFAULT_API_VERSION

        # Sanitize hostname
        if not self.host.startswith("https://"):
            if self.host.startswith("http://"):
                raise ValueError(
                    "HTTP connections are not supported, please use HTTPS instead."
                )
            else:
                self.host = "https://" + self.host

    @property
    def base_url(self) -> str:
        """
        Return the base UR for accessing the cluster.
        """
        return f"{self.host}{self.api_pattern}"

    @property
    def api_version(self) -> str:
        """
        Return the api version string to be used in the URL for making REST requests.
        """
        return self._api_version

    @api_version.setter
    def api_version(self, value: str) -> None:
        """
        Set the api_version string to be used in the URL for making REST requests.
        """
        if not value.startswith("v"):
            raise ValueError(
                "API version string should follow the format: 'vX', where X is the "
                "(integer) version number, e.g.: 'v1', 'v2', 'v3'."
            )
        self._api_version = value

    @property
    def api_pattern(self) -> str:
        """
        Return the API pattern used in the URL
        """
        return f"/api/{self.api_version}/"


@attrs.define(slots=False)
class ServerCredentialConfig(ServerConfig):
    """
    Configuration for an Intel® Geti™ server that requires authentication via username
    and password.

    NOTE: This is a legacy authentication method. Recent server versions should
    authenticate via a personal access token (API key)

    :var username: Username to log in to the instance.
    :var password: Password required to log in to the instance
    """

    username: str
    password: str


@attrs.define(slots=False)
class ServerTokenConfig(ServerConfig):
    """
    Configuration for an Intel® Geti™ server that uses a personal access token
    (API key) for authentication.

    :var token: Personal access token that can be used to connect to the server.
    """

    token: str


@attrs.define(slots=False)
class SaaSTokenConfig(ServerTokenConfig):
    """
    Configuration for the Intel® Geti™ SaaS environment that uses a personal access token
    (API key) for authentication.

    :var token: Personal access token that can be used to connect to the server.
    :var organization_id: ID of the organization for which the token is valid
    """

    organization_id: str



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import ClassVar, List, Optional, Sequence, Union

import attr

from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.shapes import Ellipse, Polygon, Rectangle, RotatedRectangle
from geti_sdk.data_models.utils import deidentify, str_to_datetime


@attr.define
class Annotation:
    """
    Representation of a single annotation for a media item on the Intel® Geti™ platform.

    :var labels: List of labels belonging to the annotation
    :var modified: Date and time of the last modification made to this annotation
    :var shape: Shape of the annotation
    :var id: Unique database ID assigned to the annotation
    :var labels_to_revisit: Optional list of database ID's of the labels that may not
        be up to date and therefore need to be revisited for this annotation
    """

    _identifier_fields: ClassVar[str] = ["id", "modified"]

    labels: List[ScoredLabel]
    shape: Union[Rectangle, Ellipse, Polygon, RotatedRectangle]
    modified: Optional[str] = attr.field(converter=str_to_datetime, default=None)
    id: Optional[str] = None
    labels_to_revisit: Optional[List[str]] = None

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the annotation and the entities it
        contains
        """
        deidentify(self)
        for label in self.labels:
            deidentify(label)

    @property
    def label_names(self) -> List[str]:
        """
        Return a list of label names for this Annotation.
        """
        return [label.name for label in self.labels]

    def append_label(self, label: ScoredLabel) -> None:
        """
        Add a label to the list of labels belonging to this annotation.

        :param label:
        :return:
        """
        self.labels.append(label)

    def extend_labels(self, labels: List[ScoredLabel]) -> None:
        """
        Add a list of labels to the labels already attached to this annotation.

        :param labels: List of ScoredLabels to add
        :return:
        """
        self.labels.extend(labels)

    def pop_label_by_name(self, label_name: str) -> None:
        """
        Remove a label from the list of labels belonging to this annotation.

        :param label_name: Name of the label to remove from the list
        :return:
        """
        index = None
        for index, label in enumerate(self.labels):
            if label.name == label_name:
                break
        if index is not None:
            self.labels.pop(index)

    def map_labels(self, labels: Sequence[Union[ScoredLabel, Label]]) -> "Annotation":
        """
        Attempt to map the labels found in `labels` to those in the Annotation
        instance. Labels are matched by name. This method will return a new
        Annotation object.

        :param labels: Labels to which the existing labels should be mapped
        :return: Annotation with updated labels, corresponding to those found in
            the `project` (if matching labels were found)
        """
        mapped_label_names = [label.name for label in labels]
        mapped_label_ids = [label.id for label in labels]

        new_labels: List[ScoredLabel] = []
        for label in self.labels:
            if label.name in mapped_label_names:
                label_index = mapped_label_names.index(label.name)
                new_labels.append(
                    ScoredLabel(
                        probability=label.probability,
                        name=label.name,
                        color=label.color,
                        id=mapped_label_ids[label_index],
                        source=label.source,
                    )
                )
        new_labels_to_revisit = [
            new_label.id
            for new_label in new_labels
            if new_label.id in self.labels_to_revisit
        ]
        return Annotation(
            labels=new_labels,
            shape=self.shape,
            modified=self.modified,
            id=None,
            labels_to_revisit=new_labels_to_revisit,
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, ClassVar, Dict, Optional

import attr

from geti_sdk.data_models import Algorithm
from geti_sdk.data_models.enums import ConfigurationEntityType
from geti_sdk.data_models.utils import attr_value_serializer, str_to_enum_converter


@attr.define
class EntityIdentifier:
    """
    Identifying information for a configurable entity on the Intel® Geti™ platform,
    as returned by the /configuration endpoint.

    :var workspace_id: ID of the workspace to which the entity belongs
    :var type: Type of the configuration
    """

    _identifier_fields: ClassVar[str] = ["workspace_id"]

    type: str = attr.field(converter=str_to_enum_converter(ConfigurationEntityType))
    workspace_id: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the EntityIdentifier.

        :return: Dictionary representing the EntityIdentifier
        """
        return attr.asdict(self, value_serializer=attr_value_serializer, recurse=True)


@attr.define
class HyperParameterGroupIdentifier(EntityIdentifier):
    """
    Identifying information for a HyperParameterGroup on the Intel® Geti™ platform,
    as returned by the /configuration endpoint.

    :var model_storage_id: ID of the model storage to which the hyper parameter group
        belongs
    :var group_name: Name of the hyper parameter group
    """

    _identifier_fields: ClassVar[str] = ["model_storage_id", "workspace_id"]

    group_name: str = attr.field(kw_only=True)
    model_storage_id: Optional[str] = None
    project_id: Optional[str] = None
    algorithm_name: Optional[str] = attr.field(repr=False, default=None)
    model_template_id: Optional[str] = attr.field(repr=False, default=None)

    def resolve_algorithm(self, algorithm: Algorithm):
        """
        Resolve the algorithm name and id of the model template to which the
        HyperParameterGroup applies.

        :param algorithm: Algorithm instance to which the hyper parameters belong
        :return:
        """
        self.algorithm_name = algorithm.name
        self.model_template_id = algorithm.model_template_id


@attr.define
class ComponentEntityIdentifier(EntityIdentifier):
    """
    Identifying information for a configurable Component on the Intel® Geti™ platform,
    as returned by the /configuration endpoint.

    :var component: Name of the component
    :var project_id: ID of the project to which the component belongs
    :var task_id: Optional ID of the task to which the component belongs
    """

    _identifier_fields: ClassVar[str] = ["project_id", "task_id", "workspace_id"]

    component: str = attr.field(kw_only=True)
    project_id: Optional[str] = None
    task_id: Optional[str] = None



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

"""Module implements the InferenceResultsToPredictionConverter class."""

import abc
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import cv2
import numpy as np
from model_api.models.utils import (
    AnomalyResult,
    ClassificationResult,
    Detection,
    DetectionResult,
    ImageResultWithSoftPrediction,
    InstanceSegmentationResult,
)

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.containers import LabelList
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.label import ScoredLabel
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.data_models.shapes import (
    Ellipse,
    Point,
    Polygon,
    Rectangle,
    RotatedRectangle,
)
from geti_sdk.deployment.predictions_postprocessing.utils.segmentation_utils import (
    create_annotation_from_segmentation_map,
)


class InferenceResultsToPredictionConverter(metaclass=abc.ABCMeta):
    """Interface for the converter"""

    def __init__(
        self, labels: LabelList, configuration: Optional[Dict[str, Any]] = None
    ):
        self.labels = labels.get_non_empty_labels()
        self.empty_label = labels.get_empty_label()
        self.configuration = configuration

    @abc.abstractmethod
    def convert_to_prediction(
        self, inference_results: NamedTuple, **kwargs
    ) -> Prediction:
        """
        Convert raw inference results to the Prediction format.

        :param inference_results: raw predictions from inference
        :return: Prediction object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def convert_saliency_map(
        self, inference_results: NamedTuple, **kwargs
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: raw predictions from inference
        :return: Prediction object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError


class ClassificationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Classification predictions to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: Optional configuration dictionary containing additional
        parameters
    """

    def __init__(
        self, labels: LabelList, configuration: Optional[Dict[str, Any]] = None
    ):
        super().__init__(labels, configuration)

    def convert_to_prediction(
        self,
        inference_results: ClassificationResult,
        image_shape: Tuple[int, int, int],
        **kwargs,
    ) -> Prediction:  # noqa: ARG003
        """
        Convert ModelAPI ClassificationResult inference results to Prediction object.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        labels = []
        for label in inference_results.top_labels:
            label_idx, label_name, label_prob = label
            # label_idx does not necessarily match the label index in the project
            # labels. Therefore, we map the label by name instead.
            labels.append(
                self.labels.create_scored_label(id_or_name=label_name, score=label_prob)
            )

        if not labels and self.empty_label:
            labels = [ScoredLabel.from_label(self.empty_label, probability=0)]

        annotations = Annotation(
            shape=Rectangle.generate_full_box(image_shape[1], image_shape[0]),
            labels=labels,
        )
        return Prediction([annotations])

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: Tuple[int, int, int],
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        saliency_map = cv2.resize(
            np.transpose(saliency_map.squeeze(0), (1, 2, 0)),
            dsize=(image_shape[1], image_shape[0]),
            interpolation=cv2.INTER_CUBIC,
        )
        if len(saliency_map.shape) == 2:
            saliency_map = np.expand_dims(saliency_map, axis=-1)
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {
            label.name: saliency_map[i]
            for i, label in enumerate(self.labels.get_non_empty_labels())
        }


class DetectionToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Detection objects to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def __init__(
        self, labels: LabelList, configuration: Optional[Dict[str, Any]] = None
    ):
        super().__init__(labels, configuration)
        self.use_ellipse_shapes = False
        self.confidence_threshold = 0.0
        if configuration is not None:
            if "use_ellipse_shapes" in configuration:
                self.use_ellipse_shapes = configuration["use_ellipse_shapes"]
            if "confidence_threshold" in configuration:
                self.confidence_threshold = configuration["confidence_threshold"]
            if "label_ids" in configuration:
                # Make sure the list of labels is sorted according to the order
                # defined in the ModelAPI configuration. If the 'label_ids' field
                # only contains a single label, it will be typed as string. No need
                # to sort in that case
                ids = configuration["label_ids"]
                if not isinstance(ids, str):
                    self.labels.sort_by_ids(configuration["label_ids"])

    def _detection2array(self, detections: List[Detection]) -> np.ndarray:
        """
        Convert list of OpenVINO Detection to a numpy array.

        :param detections: list of OpenVINO Detection containing [score, id, xmin, ymin, xmax, ymax]
        :return: numpy array with [label, confidence, x1, y1, x2, y2]
        """
        scores = np.empty((0, 1), dtype=np.float32)
        labels = np.empty((0, 1), dtype=np.uint32)
        boxes = np.empty((0, 4), dtype=np.float32)
        for det in detections:
            if (det.xmax - det.xmin) * (det.ymax - det.ymin) < 1.0:
                continue
            scores = np.append(scores, [[det.score]], axis=0)
            labels = np.append(labels, [[det.id]], axis=0)
            boxes = np.append(
                boxes,
                [[float(det.xmin), float(det.ymin), float(det.xmax), float(det.ymax)]],
                axis=0,
            )
        return np.concatenate((labels, scores, boxes), -1)

    def convert_to_prediction(
        self, inference_results: DetectionResult, **kwargs
    ) -> Prediction:
        """
        Convert ModelAPI DetectionResult inference results to Prediction object.

        :param inference_results: detection represented in ModelAPI format (label, confidence, x1, y1, x2, y2).

        _Note:
            - `label` can be any integer that can be mapped to `self.labels`
            - `confidence` should be a value between 0 and 1
            - `x1`, `x2`, `y1` and `y2` are expected to be in pixel
        :return: Prediction object containing the boxes obtained from the prediction
        """
        detections = self._detection2array(inference_results.objects)

        annotations = []
        if (
            len(detections)
            and detections.shape[1:] < (6,)
            or detections.shape[1:] > (7,)
        ):
            raise ValueError(
                f"Shape of prediction is not expected, expected (n, 7) or (n, 6) but got {detections.shape}"
            )

        for detection in detections:
            # Some OpenVINO models use an output shape of [7,]
            # If this is the case, skip the first value as it is not used
            _detection = detection[1:] if detection.shape == (7,) else detection

            label_index = int(_detection[0])
            confidence = _detection[1]
            scored_label = ScoredLabel.from_label(self.labels[label_index], confidence)
            coords = _detection[2:]
            shape: Union[Ellipse, Rectangle]

            if confidence < self.confidence_threshold:
                continue

            bbox_width = coords[2] - coords[0]
            bbox_height = coords[3] - coords[1]
            if self.use_ellipse_shapes:
                shape = Ellipse(coords[0], coords[1], bbox_width, bbox_height)
            else:
                shape = Rectangle(coords[0], coords[1], bbox_width, bbox_height)

            annotation = Annotation(shape=shape, labels=[scored_label])
            annotations.append(annotation)
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: Tuple[int, int, int],
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        if isinstance(saliency_map, list):
            saliency_map = np.array(
                [
                    smap if len(smap) > 0 else np.zeros(image_shape[:2], dtype=np.uint8)
                    for smap in saliency_map
                ]
            )
        elif isinstance(saliency_map, np.ndarray):
            saliency_map = saliency_map.squeeze(0)
        else:
            raise ValueError(
                f"Unsupported saliency map type: {type(saliency_map)}. Expected list or numpy array."
            )
        saliency_map = cv2.resize(
            np.transpose(saliency_map, (1, 2, 0)),
            dsize=(image_shape[1], image_shape[0]),
            interpolation=cv2.INTER_CUBIC,
        )
        if len(saliency_map.shape) == 2:
            saliency_map = np.expand_dims(saliency_map, axis=-1)
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in enumerate(self.labels)}


class RotatedRectToPredictionConverter(DetectionToPredictionConverter):
    """
    Converts ModelAPI Rotated Detection objects to Prediction.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def convert_to_prediction(
        self, inference_results: InstanceSegmentationResult, **kwargs
    ) -> Prediction:
        """
        Convert ModelAPI instance segmentation inference results to a rotated bounding box annotation format.

        :param inference_results: segmentation represented in ModelAPI format
        :return: Prediction object containing the rotated boxes obtained from the segmentation contours
        :raises ValueError: if metadata is missing from the preprocess step
        """
        annotations = []
        shape: Union[RotatedRectangle, Ellipse]
        for obj in inference_results.segmentedObjects:
            if obj.score < self.confidence_threshold:
                continue
            if self.use_ellipse_shapes:
                shape = Ellipse(
                    obj.xmin, obj.ymin, obj.xmax - obj.xmin, obj.ymax - obj.ymin
                )
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[
                            ScoredLabel.from_label(
                                self.labels[int(obj.id) - 1], float(obj.score)
                            )
                        ],
                    )
                )
            else:
                mask = obj.mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(
                    mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE
                )
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    points = [
                        Point(
                            x=point[0],
                            y=point[1],
                        )
                        for point in cv2.boxPoints(cv2.minAreaRect(contour))
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=RotatedRectangle.from_polygon(shape),
                            labels=[
                                ScoredLabel.from_label(
                                    self.labels[int(obj.id) - 1], float(obj.score)
                                )
                            ],
                        )
                    )
        return Prediction(annotations)


class MaskToAnnotationConverter(DetectionToPredictionConverter):
    """
    Converts DetectionBox Predictions ModelAPI to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def convert_to_prediction(
        self, inference_results: Any, **kwargs: Dict[str, Any]
    ) -> Prediction:
        """
        Convert inference results to Prediction object.

        :param inference_results: Raw inference results from the model.
        :return: Prediction object.
        """
        annotations = []
        shape: Union[Polygon, Ellipse]
        for obj in inference_results.segmentedObjects:
            if obj.score < self.confidence_threshold:
                continue
            if self.use_ellipse_shapes:
                shape = Ellipse(
                    obj.xmin, obj.ymin, obj.xmax - obj.xmin, obj.ymax - obj.ymin
                )
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[
                            ScoredLabel.from_label(
                                self.labels[int(obj.id) - 1], float(obj.score)
                            )
                        ],
                    )
                )
            else:
                mask = obj.mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(
                    mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE
                )
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    contour = list(contour)
                    points = [
                        Point(
                            x=point[0][0],
                            y=point[0][1],
                        )
                        for point in contour
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=shape,
                            labels=[
                                ScoredLabel.from_label(
                                    self.labels[int(obj.id) - 1], float(obj.score)
                                )
                            ],
                        )
                    )
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: Tuple[int, int, int],
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        if len(inference_results.saliency_map) == 0:
            return None
        # Model API returns a list of np.ndarray for each label
        # Including `no_object` which is empty
        saliency_map = np.array(
            [
                smap if len(smap) > 0 else np.zeros(image_shape[:2], dtype=np.uint8)
                for smap in inference_results.saliency_map
            ]
        )  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in enumerate(self.labels)}


class SegmentationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Segmentation objects to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def __init__(
        self, labels: LabelList, configuration: Optional[Dict[str, Any]] = None
    ):
        super().__init__(labels, configuration)
        # NB: index=0 is reserved for the background label
        self.label_map = dict(enumerate(self.labels, 1))

    def convert_to_prediction(
        self, inference_results: ImageResultWithSoftPrediction, **kwargs  # noqa: ARG002
    ) -> Prediction:
        """
        Convert ModelAPI instance segmentation inference results to Prediction object.

        :param inference_results: segmentation represented in ModelAPI format
        :return: Prediction object containing the contour polygon obtained from the segmentation
        """
        annotations = create_annotation_from_segmentation_map(
            hard_prediction=inference_results.resultImage,
            soft_prediction=inference_results.soft_prediction,
            label_map=self.label_map,
        )
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: Tuple[int, int, int],
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {label.name: saliency_map[i + 1] for i, label in enumerate(self.labels)}


class AnomalyToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Convert ModelAPI AnomalyResult predictions to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def __init__(
        self, labels: LabelList, configuration: Optional[Dict[str, Any]] = None
    ):
        super().__init__(labels, configuration)
        self.normal_label = next(
            label for label in self.labels if not label.is_anomalous
        )
        self.anomalous_label = next(
            label for label in self.labels if label.is_anomalous
        )
        if configuration is not None and "domain" in configuration:
            self.domain = configuration["domain"]

    def convert_to_prediction(
        self, inference_results: AnomalyResult, image_shape: Tuple[int], **kwargs
    ) -> Prediction:  # noqa: ARG002
        """
        Convert ModelAPI AnomalyResult inferenceresults to sc_sdk annotations.

        :param inference_results: anomaly result represented in ModelAPI format (same for all anomaly tasks)
        :return: Prediction object based on the specific anomaly task:
            - Classification: single label (normal or anomalous).
            - Segmentation: contour polygon representing the segmentation.
            - Detection: predicted bounding boxes.
        """
        pred_label = inference_results.pred_label
        label = (
            self.anomalous_label
            if pred_label in ("Anomaly", "Anomalous")
            else self.normal_label
        )
        annotations: List[Annotation] = []
        if (
            self.domain == Domain.ANOMALY_CLASSIFICATION
            or self.domain == Domain.ANOMALY
        ):
            scored_label = ScoredLabel.from_label(
                label=label, probability=float(inference_results.pred_score)
            )
            annotations = [
                Annotation(
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                    labels=[scored_label],
                )
            ]
        elif self.domain == Domain.ANOMALY_SEGMENTATION:
            annotations = create_annotation_from_segmentation_map(
                hard_prediction=inference_results.pred_mask,
                soft_prediction=inference_results.anomaly_map.squeeze(),
                label_map={0: self.normal_label, 1: self.anomalous_label},
            )
        elif self.domain == Domain.ANOMALY_DETECTION:
            for box in inference_results.pred_boxes:
                annotations.append(
                    Annotation(
                        shape=Rectangle(
                            box[0], box[1], box[2] - box[0], box[3] - box[1]
                        ),
                        labels=[
                            ScoredLabel.from_label(
                                label=self.anomalous_label,
                                probability=inference_results.pred_score,
                            )
                        ],
                    )
                )
        else:
            raise ValueError(
                f"Cannot convert inference results for task '{self.domain.name}'. Only Anomaly tasks are supported."
            )
        if not annotations:
            scored_label = ScoredLabel.from_label(
                label=self.normal_label, probability=0
            )
            annotations = [
                Annotation(
                    labels=[scored_label],
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                )
            ]
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: Tuple[int, int, int],
    ) -> Optional[Dict[str, np.ndarray]]:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        # Normalizing Anomaly map
        saliency_map = inference_results.anomaly_map
        saliency_map -= saliency_map.min()
        saliency_map = saliency_map / (saliency_map.max() + 1e-12) * 255
        saliency_map = np.round(saliency_map).astype(np.uint8)  # shape: (h, w)
        return {self.anomalous_label.name: saliency_map}


class ConverterFactory:
    """
    Factory class for creating inference result to prediction converters based on the model's task.
    """

    @staticmethod
    def create_converter(
        labels: LabelList,
        domain: Domain,
        configuration: Optional[Dict[str, Any]] = None,
    ) -> InferenceResultsToPredictionConverter:
        """
        Create the appropriate inferencer object according to the model's task.

        :param label_schema: The label schema containing the label info of the task.
        :param domain: The domain to which the converter applies
        :param configuration: Optional configuration for the converter. Defaults to None.
        :return: The created inference result to prediction converter.
        :raises ValueError: If the task type cannot be determined from the label schema.
        """
        if domain == Domain.CLASSIFICATION:
            return ClassificationToPredictionConverter(labels, configuration)
        if domain == Domain.DETECTION:
            return DetectionToPredictionConverter(labels, configuration)
        if domain == Domain.SEGMENTATION:
            return SegmentationToPredictionConverter(labels, configuration)
        if domain == Domain.ROTATED_DETECTION:
            return RotatedRectToPredictionConverter(labels, configuration)
        if domain == Domain.INSTANCE_SEGMENTATION:
            return MaskToAnnotationConverter(labels, configuration)
        if domain in (
            Domain.ANOMALY_CLASSIFICATION,
            Domain.ANOMALY_SEGMENTATION,
            Domain.ANOMALY_DETECTION,
            Domain.ANOMALY,
        ):
            configuration.update({"domain": domain})
            return AnomalyToPredictionConverter(labels, configuration)
        raise ValueError(f"Cannot create inferencer for task type '{domain.name}'.")



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from typing import List, Optional, Sequence

from geti_sdk.data_models import Dataset, Job, Model, Project, TestResult
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import TestResultRESTConverter
from geti_sdk.utils.job_helpers import get_job_with_timeout, monitor_job, monitor_jobs

SUPPORTED_METRICS = ["global", "local"]


class TestingClient:
    """
    Class to manage testing jobs for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project.id}/tests"

    def test_model(
        self,
        model: Model,
        datasets: Sequence[Dataset],
        name: Optional[str] = None,
        metric: Optional[str] = None,
    ) -> Job:
        """
        Start a model testing job for a specific `model` and `dataset`

        :param model: Model to evaluate
        :param datasets: Testing dataset(s) to evaluate the model on
        :param name: Optional name to assign to the testing job
        :param metric: Optional metric to calculate. This is only valid for either
            anomaly segmentation or anomaly detection models. Possible values are
            `global` or `local`
        :return: Job object representing the testing job
        """
        if name is None:
            name = f"Test_{model.name}_{model.version}"
        dataset_ids = [ds.id for ds in datasets]

        test_data = {
            "name": name,
            "model_group_id": model.model_group_id,
            "model_id": model.id,
            "dataset_ids": dataset_ids,
        }
        if metric is not None:
            if metric not in SUPPORTED_METRICS:
                raise ValueError(
                    f"Invalid metric received! Only `{SUPPORTED_METRICS}` are "
                    f"supported currently."
                )
            test_data.update({"metric": metric})

        response = self.session.get_rest_response(
            url=self.base_url, method="POST", data=test_data
        )
        job = get_job_with_timeout(
            job_id=response["job_id"],
            session=self.session,
            workspace_id=self.workspace_id,
            job_type="testing",
        )
        logging.info(f"Testing job with id {job.id} submitted successfully.")
        return job

    def get_test_result(self, job: Job) -> TestResult:
        """
        Retrieve the result of the model testing job from the Intel® Geti™
        server

        :param job: Job instance representing the model testing job
        :return: TestResult instance containing the test results
        """
        response = self.session.get_rest_response(url=self.base_url, method="GET")
        test_results = [
            TestResultRESTConverter.from_dict(result)
            for result in response["test_results"]
        ]
        result_for_job = next(
            (result for result in test_results if result.job_info.id == job.id), None
        )
        if result_for_job is None:
            raise ValueError(
                f"Unable to find test result for job `{job.name}`, please make sure "
                f"that a valid testing job was passed."
            )
        return result_for_job

    def monitor_jobs(
        self, jobs: List[Job], timeout: int = 10000, interval: int = 15
    ) -> List[Job]:
        """
        Monitor and print the progress of all jobs in the list `jobs`. Execution is
        halted until all jobs have either finished, failed or were cancelled.

        Progress will be reported in 15s intervals

        :param jobs: List of jobs to monitor
        :param timeout: Timeout (in seconds) after which to stop the monitoring
        :param interval: Time interval (in seconds) at which the TrainingClient polls
            the server to update the status of the jobs. Defaults to 15 seconds
        :return: List of finished (or failed) jobs with their status updated
        """
        return monitor_jobs(
            session=self.session, jobs=jobs, timeout=timeout, interval=interval
        )

    def monitor_job(self, job: Job, timeout: int = 10000, interval: int = 15) -> Job:
        """
        Monitor and print the progress of the `job`. Execution is
        halted until the job has either finished, failed or was cancelled.

        Progress will be reported in 15s intervals

        :param job: Job to monitor
        :param timeout: Timeout (in seconds) after which to stop the monitoring
        :param interval: Time interval (in seconds) at which the TrainingClient polls
            the server to update the status of the jobs. Defaults to 15 seconds
        :return: Job with its status updated
        """
        return monitor_job(
            session=self.session, job=job, timeout=timeout, interval=interval
        )



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from geti_sdk.data_models import Image, Project, Video
from geti_sdk.data_models.containers import MediaList
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import MediaRESTConverter


class ActiveLearningClient:
    """
    Class to manage the active learning for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = (
            f"workspaces/{workspace_id}/projects/{project.id}/datasets/active"
        )

    def get_active_set(self) -> MediaList:
        """
        Retrieve the active dataset for the project

        :return: MediaList containing the images and videoframes that are in the
            current active set for the project
        """
        result = self.session.get_rest_response(url=self.base_url, method="GET")
        active_set = MediaList([])
        for media_item in result["active_set"]:
            media_item.pop("dataset_id")
            if media_item["type"] == "image":
                active_set.append(
                    MediaRESTConverter.from_dict(media_item, media_type=Image)
                )
            elif media_item["type"] == "video":
                active_frames = media_item.pop("active_frames")
                video = MediaRESTConverter.from_dict(media_item, media_type=Video)
                frame_list = [video.get_frame(frame_index=i) for i in active_frames]
                active_set.extend(frame_list)
        return active_set



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import os
from typing import Optional, Union

from dotenv import dotenv_values

from geti_sdk.http_session.server_config import (
    ServerCredentialConfig,
    ServerTokenConfig,
)


def convert_boolean_env_variable(
    value: Optional[str], default_value: bool = False
) -> bool:
    """
    Convert a string that was extracted from environment variables to a
    boolean.

    Strings like 'True', 'true', 't', '1', 'T' are accepted and converted to True
    Strings like 'False', 'false', 'f', '0', 'F', are accepted and converted to False
    Other strings are rejected, a ValueError will be raised in that case

    :param value: string to convert to boolean
    :param default_value: Default value to return, will be returned in case `value` is
        None
    :return: Boolean corresponding to the input `value`
    """
    true_strings = ("true", "1", "t")
    false_strings = ("false", "0", "f")

    if value is None:
        return default_value

    if value.lower() not in true_strings + false_strings:
        raise ValueError(f"Invalid value `{value}` for boolean variable")
    return value in true_strings


def get_server_details_from_env(
    env_file_path: str = ".env", use_global_variables: bool = False
) -> Union[ServerTokenConfig, ServerCredentialConfig]:
    """
    Retrieve the server information (hostname and authentication details) from
    environment variables.

    By default, this method will attempt to find a file ".env" in the current working
    directory and attempt to read the information from that file. To read from a
    different file, simply specify the path to the file in `env_file_path`. The
    following variables are relevant:

        HOST        -> hostname or ip address of the Geti server

        TOKEN       -> Personal Access Token that can be used for authorization

        VERIFY_CERT -> boolean, pass 1 or True to verify

        HTTPS_PROXY -> Address of the proxy to be used for https communication. Make
                        sure to specify the full address, including port number. For
                        example: HTTPS_PROXY=http://proxy.my-company.com:900

    In addition, authentication via credentials is also supported. In that case, the
    following variables should be provided:

        USERNAME -> username to log in to the Geti server

        PASSWORD -> password for logging in to the Geti server


    If both TOKEN, USERNAME and PASSWORD are provided, the method will use the preferred
    token authorization.

    If `use_global_variables` is set to `True`, the method attempts to retrieve the
    server details from the global environment variables instead of
    reading from file. In that case, the `GETI_` prefix should be added to the
    variable names (i.e. `HOST` becomes `GETI_HOST`, `TOKEN` becomes `GETI_TOKEN`, etc.)

    :param env_file_path: Path to the file containing the server details
    :param use_global_variables: If set to True, the method will not read the server
        details from a file but will use environment variables instead
    :return: A ServerConfig instance that contains the details of the Geti server
        specified in the environment
    """
    if use_global_variables:
        host_key = "GETI_HOST"
        token_key = "GETI_TOKEN"  # nosec: B105
        username_key = "GETI_USERNAME"
        password_key = "GETI_PASSWORD"  # nosec: B105
        cert_key = "GETI_VERIFY_CERT"
        https_proxy_key = "GETI_HTTPS_PROXY"

        retrieval_func = os.environ.get
        env_name = "environment variables"
    else:
        host_key = "HOST"
        token_key = "TOKEN"  # nosec: B105
        username_key = "USERNAME"
        password_key = "PASSWORD"  # nosec: B105
        cert_key = "VERIFY_CERT"
        https_proxy_key = "HTTPS_PROXY"

        env_variables = dotenv_values(dotenv_path=env_file_path)
        if not env_variables:
            raise ValueError(
                f"Unable to load Geti server details from environment file at path "
                f"'{env_file_path}', please make sure the file exists."
            )
        retrieval_func = env_variables.get
        env_name = "environment file"

    # Extract hostname
    hostname = retrieval_func(host_key, None)
    if hostname is None:
        raise ValueError(
            f"Unable to find Geti hostname in the {env_name}. Please make sure "
            f"that the variable {host_key} is defined."
        )

    # Extract certificate validation. Defaults to True
    verify_certificate = convert_boolean_env_variable(
        value=retrieval_func(cert_key, None), default_value=True
    )

    # Extract https proxy configuration
    https_proxy = retrieval_func(https_proxy_key, None)
    if https_proxy is not None:
        proxies = {"https": https_proxy}
    else:
        proxies = None

    # Extract token/credentials
    token = retrieval_func(token_key, None)
    if token is None:
        user = retrieval_func(username_key, None)
        password = retrieval_func(password_key, None)
        if user is None or password is None:
            raise ValueError(
                f"Unable to find either personal access token or user credentials in "
                f"the {env_name}. Please make sure that either '{token_key}' or "
                f"'{username_key}' and '{password_key}' is defined in the environment."
            )
        server_config = ServerCredentialConfig(
            host=hostname,
            username=user,
            password=password,
            has_valid_certificate=verify_certificate,
            proxies=proxies,
        )
    else:
        server_config = ServerTokenConfig(
            host=hostname,
            token=token,
            has_valid_certificate=verify_certificate,
            proxies=proxies,
        )
    return server_config



import os
import time
from io import BufferedReader
from typing import Optional

from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.http_session.geti_session import GetiSession


class TUSUploader:
    """
    Class to handle tus uploads.
    """

    DEFAULT_CHUNK_SIZE = 5 * 2**20  # 5MB

    def __init__(
        self,
        session: GetiSession,
        base_url: str,
        file_path: Optional[os.PathLike] = None,
        file_stream: Optional[BufferedReader] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        retries: int = 0,
        retry_delay: int = 30,
    ):
        """
        Initialize TUSUploader instance.

        :param session: GetiSession instance.
        :param base_url: Base url for the tus upload.
        :param file_path: Path to the file to be uploaded.
        :param file_stream: File stream of the file to be uploaded.
        :param chunk_size: Size of the chunk to be uploaded at each cycle.
        :param retries: Number of retries to be made in case of upload failure.
        :param retry_delay: Delay between retries.
        """
        if file_path is None and file_stream is None:
            raise ValueError("Either 'file_path' or 'file_stream' cannot be None.")

        self.file_path = file_path
        self.file_stream = file_stream
        if self.file_stream is None:
            self.file_stream = self.get_file_stream()
        self.stop_at = self.get_file_size()
        self.session = session
        self.base_url = base_url
        self.tus_resumable_version = self._get_tus_resumable_version()
        self.offset = 0
        self.upload_url = None
        self.chunk_size = chunk_size
        self.retries = retries
        self.request = None
        self._retried = 0
        self.retry_delay = retry_delay

    def _get_tus_resumable_version(self):
        """
        Return tus resumable version.
        """
        response = self.session.get_rest_response(
            url=self.base_url,
            method="OPTIONS",
        )
        return response.headers["tus-resumable"]

    def get_offset(self):
        """
        Return offset from tus server.

        Make an http request to the tus server to retrieve the current offset.

        :return: Offset value.
        :raises: Exception if offset retrieval fails.
        """
        response = self.session.get_rest_response(
            url=self.upload_url,
            method="HEAD",
            request_headers={
                "tus-resumable": self.tus_resumable_version,
            },
        )
        offset = response.headers.get("upload-offset")
        if offset is None:
            raise Exception("Attempt to retrieve offset failed")
        return int(offset)

    def get_request_length(self):
        """
        Return length of next chunk upload.
        """
        remainder = self.stop_at - self.offset
        return self.chunk_size if remainder > self.chunk_size else remainder

    def get_file_stream(self) -> BufferedReader:
        """
        Return a file stream instance of the upload.

        :return: File stream instance.
        :raises: ValueError if file_path is invalid.
        """
        if self.file_stream:
            self.file_stream.seek(0)
            return self.file_stream
        elif self.file_path is not None and os.path.isfile(self.file_path):
            return open(self.file_path, "rb")
        else:
            raise ValueError("invalid file {}".format(self.file_path))

    def get_file_size(self):
        """
        Return size of the file.
        """
        stream = self.get_file_stream()
        stream.seek(0, os.SEEK_END)
        return stream.tell()

    def upload(self, stop_at: Optional[int] = None):
        """
        Perform file upload.

        Performs continous upload of chunks of the file. The size uploaded at each cycle is
        the value of the attribute 'chunk_size'.

        :param stop_at: Offset value at which the upload should stop. If not specified this
            defaults to the file size.
        """
        self.stop_at = stop_at or self.get_file_size()

        if not self.upload_url:
            self.upload_url = self.create_upload_url()
            self.offset = 0

        self.file_stream = self.get_file_stream()
        with logging_redirect_tqdm(tqdm_class=tqdm):
            with tqdm(
                total=self.stop_at >> 20,
                desc="Uploading file",
                unit="MB",
            ) as tbar:
                while self.offset < self.stop_at:
                    self.upload_chunk()
                    tbar.update((self.offset >> 20) - tbar.n)

    def create_upload_url(self):
        """
        Return upload url.

        Makes request to tus server to create a new upload url for the required file upload.
        """
        response = self.session.get_rest_response(
            url=self.base_url,
            method="POST",
            request_headers={
                "tus-resumable": self.tus_resumable_version,
                "upload-length": str(self.get_file_size()),
            },
        )
        upload_url = response.headers.get("location")
        if upload_url is None:
            raise ValueError("Upload url not returned by server")
        return upload_url

    def get_file_id(self) -> Optional[str]:
        """
        Return file id from upload url.

        :return: File id.
        """
        if (
            self.upload_url is None
            or len(file_id := self.upload_url.split("/")[-1]) < 2
        ):
            # We get the file_id from the upload url. If the url is not set or the file_id
            # is not valid (may be an empty string if the url is not valid), we return None.
            return
        return file_id

    def upload_chunk(self):
        """
        Upload chunk of file.
        """
        self._retried = 0
        try:
            self.offset = self._patch()
        except Exception as err:
            self._retry(err)

    def _patch(self) -> int:
        """
        Perform actual request.

        :return: Offset value after the request.
        """
        chunk = self.file_stream.read(self.get_request_length())
        # self.add_checksum(chunk)
        response = self.session.get_rest_response(
            url=self.upload_url,
            method="PATCH",
            data=chunk,
            contenttype="offset+octet-stream",
            request_headers={
                "upload-offset": str(self.offset),
                "tus-resumable": self.tus_resumable_version,
            },
        )
        upload_offset = int(response.headers.get("upload-offset"))
        return int(upload_offset)

    def _retry(self, error):
        """
        Retry upload in case of failure.

        :param error: Error that caused the upload to fail.
        :raises: error if retries are exhausted.
        """
        if self.retries > self._retried:
            time.sleep(self.retry_delay)

            self._retried += 1
            try:
                self.offset = self.get_offset()
            except Exception as err:
                self._retry(err)
            else:
                self.upload()
        else:
            raise error



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `data_models` package contains the SDK representation for all entities in GETi, such
as :py:class:`~geti_sdk.data_models.annotation_scene.AnnotationScene`,
:py:class:`~geti_sdk.data_models.media.Image`,
:py:class:`~geti_sdk.data_models.project.Project` and
:py:class:`~geti_sdk.data_models.model.Model` and many more.

When interacting with the GETi cluster through the
:py:class:`geti_sdk.geti.Geti` or the
:py:mod:`~geti_sdk.rest_clients`, all entities retrieved from the cluster will be
deserialized into the data models defined in this package.

Module contents
---------------

Algorithm-related entities
++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.algorithms
   :members:
   :undoc-members:

Project-related entities
++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.task
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.project
   :members:
   :undoc-members:

Credit System-related entities
++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.credit_system
   :members:
   :undoc-members:

Annotation-related entities
+++++++++++++++++++++++++++
.. automodule:: geti_sdk.data_models.label
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.shapes
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.annotations
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.annotation_scene
   :members:
   :undoc-members:

Configuration-related entities
++++++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.configurable_parameter
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configurable_parameter_group
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configuration_identifiers
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configuration
   :members:
   :undoc-members:

Model-related entities
++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.model
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.model_group
   :members:
   :undoc-members:

Media-related entities
++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.media
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.media_identifiers
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.task_annotation_state
   :members:
   :undoc-members:

Prediction-related entities
+++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.predictions
   :members:
   :undoc-members:

Status- and job-related entities
++++++++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.status
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.job
   :members:
   :undoc-members:

Deployment-related entities
+++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_model.code_deployment_info
   :members:
   :undoc-members:

Utility functions
+++++++++++++++++

.. automodule:: geti_sdk.data_models.utils
   :members:
   :undoc-members:

Custom container classes
++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.containers
   :members:
   :undoc-members:

Enumerations
++++++++++++

.. automodule:: geti_sdk.data_models.enums
   :members:
   :undoc-members:

"""

from .algorithms import Algorithm
from .annotation_scene import AnnotationScene
from .annotations import Annotation
from .code_deployment_info import CodeDeploymentInformation
from .configuration import (
    ConfigurableParameters,
    FullConfiguration,
    GlobalConfiguration,
    TaskConfiguration,
)
from .credit_system import CreditAccount, CreditBalance, Subscription
from .dataset import Dataset, Subset, TrainingDatasetStatistics
from .enums import AnnotationKind, MediaType, TaskType
from .job import Job
from .label import Label, ScoredLabel
from .media import Image, MediaItem, Video, VideoFrame
from .model import Model, OptimizedModel
from .model_group import ModelGroup, ModelSummary
from .performance import Performance
from .predictions import Prediction
from .project import Pipeline, Project
from .status import ProjectStatus
from .task import Task
from .test_result import Score, TestResult
from .user import User

__all__ = [
    "TaskType",
    "AnnotationKind",
    "Project",
    "Label",
    "Task",
    "Pipeline",
    "Dataset",
    "Image",
    "Video",
    "MediaItem",
    "MediaType",
    "Annotation",
    "AnnotationScene",
    "Algorithm",
    "ScoredLabel",
    "VideoFrame",
    "Prediction",
    "Performance",
    "TaskConfiguration",
    "GlobalConfiguration",
    "ConfigurableParameters",
    "FullConfiguration",
    "Model",
    "ModelGroup",
    "OptimizedModel",
    "ModelSummary",
    "ProjectStatus",
    "Job",
    "CodeDeploymentInformation",
    "TestResult",
    "Score",
    "User",
    "CreditAccount",
    "CreditBalance",
    "Subscription",
    "Subset",
    "TrainingDatasetStatistics",
]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import time
from multiprocessing import Value
from queue import Empty
from threading import Thread
from typing import Any, Callable, Optional, Tuple

import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment import Deployment

from .ordered_buffer import OrderedResultBuffer


class AsyncVideoProcessor:
    """
    A helper class for efficient and optimized frame-by-frame inference on videos
    """

    def __init__(
        self,
        deployment: Deployment,
        processing_function: Callable[[np.ndarray, Prediction, Any], None],
        min_buffer_size: Optional[int] = None,
        max_buffer_size: Optional[int] = None,
    ):
        """
        Run efficient and optimized frame-by-frame inference on videos. The
        AsyncVideoProcessor runs inference asynchronously,
        maximizing the host systems potential for parallelization. However, the
        internal buffer ensures that the video frame and inference result
        postprocessing is done in the original frame order.

        :param deployment: Model deployment to use for frame inference
        :param processing_function: A user defined function for further processing
            the video frames and the predictions. The function should take the
            following arguments:
                - image: np.ndarry containing the original image/video frame
                - prediction: Prediction object containing the model predictions for
                    the image/video frame
                - runtime_data: Any additional data that is required for the further
                    processing. Can be for example a filename, frame index, etc.
        :param min_buffer_size: Minimal number of frames to keep in the internal buffer
        :param max_buffer_size: Maximum number of frames that the buffer can hold.
            This can be used to limit the memory usage of the VideoProcessor
        """
        self.deployment = deployment
        self.processing_function = processing_function

        # Try to set min buffer size based on cpu count
        if min_buffer_size is None and os.cpu_count() is not None:
            min_buffer_size = 2 * os.cpu_count()
            logging.debug(
                f"Minimum buffer size set to `{min_buffer_size}`, twice the "
                f"number of available cpu's on the system."
            )
        elif min_buffer_size is not None and os.cpu_count() is not None:
            if min_buffer_size < os.cpu_count():
                logging.warning(
                    "The minimum buffer size is less than the cpu count of the system. "
                    "This may result in video frames not being processed in order. It"
                    "is recommended to increase the minimum buffer size."
                )
        # Make sure min_buffer_size is not larger than max_buffer_size
        if min_buffer_size is not None and max_buffer_size is not None:
            target_min_buffer_size = min_buffer_size
            min_buffer_size = max(2, min(target_min_buffer_size, max_buffer_size - 1))
            if min_buffer_size < target_min_buffer_size:
                logging.info(
                    f"Minimum buffer size `{target_min_buffer_size}` was set larger "
                    f"than maximum size `{max_buffer_size}`. Limiting minimum buffer "
                    f"size to {min_buffer_size}."
                )
        # Ultimately, fall back to 2 as absolute minimum
        if min_buffer_size is None or min_buffer_size < 2:
            logging.info("Setting minimum buffer size to 2")
            min_buffer_size = 2

        self.buffer = OrderedResultBuffer(
            maxsize=max_buffer_size, minsize=min_buffer_size
        )
        self._worker: Optional[Thread] = None
        self._should_stop: bool = False
        self._should_stop_now: bool = False
        self._is_running: bool = False
        self._min_buffer_size = min_buffer_size
        self._current_index = Value("i", 0)
        self._processed_count = Value("i", 0)

        if deployment.asynchronous_mode:
            logging.info(
                "Deployment is already in asynchronous mode, any previously defined "
                "callback will be overwritten by the AsyncVideoProcessor"
            )

        if not deployment.are_models_loaded:
            logging.info(
                "Inference models are not loaded, configuring them for maximal "
                "throughput and loading to CPU now."
            )
            deployment.load_inference_models(
                device="CPU",
                max_async_infer_requests=os.cpu_count(),
                openvino_configuration={"PERFORMANCE_HINT": "THROUGHPUT"},
            )

        def infer_callback(
            image: np.ndarray, prediction: Prediction, runtime_data: Tuple[int, Any]
        ):
            """
            Infer callback to put the image, prediction and runtime data into the
            ordered buffer.
            """
            index, runtime_data = runtime_data
            self.buffer.put(
                index=index,
                image=image,
                prediction=prediction,
                runtime_data=runtime_data,
            )

        self.deployment.set_asynchronous_callback(infer_callback)

    def start(self, num_frames: Optional[int] = None):
        """
        Start the processing thread.

        :param num_frames: Optional integer specifying the length of the video to
            process. When processing a continuous stream, leave this as None
        """
        if self._is_running:
            raise ValueError("The processing thread is already running.")

        # Reset the index
        with self._current_index.get_lock():
            self._current_index.value = 0

        # Reset the processing count
        with self._processed_count.get_lock():
            self._processed_count.value = 0

        def process_items():
            """
            Process an item from the buffer
            """
            while True:
                if self._should_stop_now:
                    logging.debug(
                        "AsyncVideoProcessor: Stopping processing thread immediately"
                    )
                    return

                # Check if all frames have been inferred and put to the buffer already
                all_frames_inferred = False
                if num_frames is not None:
                    all_frames_inferred = (
                        self.buffer.total_items_buffered == num_frames - 1
                    )

                empty_buffer = self._should_stop or all_frames_inferred
                # Get the item from the buffer
                try:
                    item = self.buffer.get(empty_buffer=empty_buffer, timeout=1e-6)
                except Empty:
                    with (
                        self._processed_count.get_lock(),
                        self._current_index.get_lock(),
                    ):
                        if (
                            self._processed_count.value == self._current_index.value
                            and self._should_stop
                        ):
                            # Buffer is empty, all items have been processed.
                            # Stop thread
                            return
                    time.sleep(1e-12)
                    continue

                self.processing_function(item.image, item.prediction, item.runtime_data)
                with self._processed_count.get_lock():
                    self._processed_count.value += 1

        self._worker = Thread(target=process_items, daemon=False)
        self._worker.start()
        self._is_running = True
        logging.debug("AsyncVideoProcessor: Processing thread started")

    def stop(self):
        """
        Stop the processing thread immediately. Any items left in the buffer will not
        be processed.
        """
        self._should_stop_now = True
        self._worker.join()
        self._is_running = False
        self._should_stop_now = False
        logging.debug("AsyncVideoProcessor: Processing thread stopped")

    def process(self, frame: np.ndarray, runtime_data: Optional[Any] = None):
        """
        Run inference for the frame and process the results according to the
        `processing_function` defined in the AsyncVideoProcessor initialization.

        :param frame: Numpy array representing the frame, in RGB-channel order
        :param runtime_data: Additional optional data pertaining to the frame, which
            should be passed to the processing_function
        """
        if not self._is_running:
            raise ValueError(
                "The processing thread is not running, please start it using the "
                "`.start()` method first."
            )
        with self._current_index.get_lock():
            self.deployment.infer_async(
                image=frame, runtime_data=(self._current_index.value, runtime_data)
            )
            self._current_index.value += 1

    def await_all(self):
        """
        Block program execution until all frames left in the buffer have been fully
        processed.
        """
        if not self._is_running:
            return
        self._should_stop = True
        logging.info("AsyncVideoProcessor: Awaiting processing of all frames in buffer")
        self._worker.join()
        self._is_running = False
        self._should_stop = False
        logging.info(
            "AsyncVideoProcessor: All frames processed. Processing thread stopped"
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional, Union

import cv2
import numpy as np
from IPython.display import display
from PIL import Image as PILImage

from geti_sdk.data_models.annotation_scene import AnnotationScene
from geti_sdk.data_models.containers import MediaList
from geti_sdk.data_models.enums import AnnotationKind
from geti_sdk.data_models.media import Image, VideoFrame
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.prediction_visualization.visualizer import Visualizer


def show_image_with_annotation_scene(
    image: Union[Image, VideoFrame, np.ndarray],
    annotation_scene: Union[AnnotationScene, Prediction],
    filepath: Optional[str] = None,
    show_in_notebook: bool = False,
    show_results: bool = True,
    channel_order: str = "rgb",
    show_labels: bool = True,
    show_confidences: bool = True,
    fill_shapes: bool = True,
) -> np.ndarray:
    """
    Display an image with an annotation_scene overlayed on top of it.

    :param image: Image to show prediction for.
    :param annotation_scene: Annotations or Predictions to overlay on the image
    :param filepath: Optional filepath to save the image with annotation overlay to.
        If left as None, the result will not be saved to file
    :param show_in_notebook: True if the image needs to be shown in a notebook context.
        Setting this to True will display the image inline in the notebook. Setting it
        to False will open a pop up to show the image.
    :param show_results: True to show the results. If `show_in_notebook` is True, this
        will display the image with the annotations inside the notebook. If
        `show_in_notebook` is False, a new opencv window will pop up. If
        `show_results` is set to False, the results will not be shown but will only
        be returned instead
    :param channel_order: The channel order (R,G,B or B,G,R) used for the input image.
        This parameter accepts either `rgb` or `bgr` as input values, and defaults to
        `rgb`.
    :param show_labels: True to show the labels of the annotations. If set to False,
        the labels will not be shown.
    :param show_confidences: True to show the confidences of the annotations. If set to
        False, the confidences will not be shown.
    :param fill_shapes: True to fill the shapes of the annotations. If set to False, the
        shapes will not be filled.
    """
    if annotation_scene.kind == AnnotationKind.ANNOTATION:
        plot_type = "Annotation"
    elif annotation_scene.kind == AnnotationKind.PREDICTION:
        plot_type = "Prediction"
    else:
        raise ValueError(
            f"Invalid input: Unable to plot object of type {type(annotation_scene)}."
        )
    if isinstance(image, np.ndarray):
        name = "Numpy image"
    else:
        name = image.name

    window_name = f"{plot_type} for {name}"
    visualizer = Visualizer(
        window_name=window_name,
        show_labels=show_labels,
        show_confidence=show_confidences,
    )

    if isinstance(image, np.ndarray):
        numpy_image = image.copy()
    else:
        numpy_image = image.numpy.copy()

    if channel_order == "bgr":
        rgb_image = cv2.cvtColor(numpy_image, cv2.COLOR_BGR2RGB)
    elif channel_order == "rgb":
        rgb_image = numpy_image
    else:
        raise ValueError(
            f"Invalid channel order '{channel_order}'. Please use either `rgb` or "
            f"`bgr`."
        )

    result = visualizer.draw(
        image=rgb_image, annotation=annotation_scene, fill_shapes=fill_shapes
    )

    # For compatibility with the previous version of the function
    # return image in BGR order; to be changed in 2.0.
    result_bgr = cv2.cvtColor(result, cv2.COLOR_RGB2BGR)

    if filepath is None:
        if show_results:
            image = PILImage.fromarray(result)
            if not show_in_notebook:
                image.show(title=window_name)
            else:
                display(image)
    else:
        success, buffer = cv2.imencode(".jpg", result_bgr)
        if success:
            buffer.tofile(filepath)
        else:
            raise RuntimeError("Unable to encode output image to .jpg format.")

    return result_bgr


def show_video_frames_with_annotation_scenes(
    video_frames: MediaList[VideoFrame],
    annotation_scenes: List[Union[AnnotationScene, Prediction]],
    wait_time: float = 1,
    filepath: Optional[str] = None,
    show_labels: bool = True,
    show_confidences: bool = True,
    fill_shapes: bool = True,
):
    """
    Display a list of VideoFrames, with their annotations or predictions overlayed on
    top. The parameter `wait_time` specifies the time each frame is shown, in seconds.

    :param video_frames: List of VideoFrames to show
    :param annotation_scenes: List of AnnotationsScenes or Predictions to overlay on
        the frames
    :param wait_time: Time to show each frame, in seconds
    :param filepath: Optional filepath to save the video with annotation overlay to.
        If left as None, the video frames will be shown in a new opencv window
    :param show_labels: True to show the labels of the annotations. If set to False,
        the labels will not be shown.
    :param show_confidences: True to show the confidences of the annotations. If set to
        False, the confidences will not be shown.
    :param fill_shapes: True to fill the shapes of the annotations. If set to False, the
        shapes will not be filled.
    """
    first_frame = video_frames[0]

    out_writer: Optional[cv2.VideoWriter] = None
    if filepath is not None:
        out_writer = cv2.VideoWriter(
            filename=f"{filepath}",
            fourcc=cv2.VideoWriter_fourcc("M", "J", "P", "G"),
            fps=1 / wait_time,
            frameSize=(
                first_frame.media_information.width,
                first_frame.media_information.height,
            ),
        )

    if annotation_scenes[0].kind == AnnotationKind.ANNOTATION:
        name = "Annotation"
    elif annotation_scenes[0].kind == AnnotationKind.PREDICTION:
        name = "Prediction"
    else:
        raise ValueError(
            f"Invalid input: Unable to plot object of type "
            f"{type(annotation_scenes[0])}."
        )
    window_name = f"{name} for '{video_frames[0].video_name}'"
    visualizer = Visualizer(
        window_name=window_name,
        show_labels=show_labels,
        show_confidence=show_confidences,
    )

    for frame, annotation_scene in zip(video_frames, annotation_scenes):
        numpy_frame = frame.numpy.copy()
        result = visualizer.draw(
            numpy_frame, annotation=annotation_scene, fill_shapes=fill_shapes
        )

        if out_writer is None:
            cv2.imshow(window_name, result)
            cv2.waitKey(int(wait_time * 1000))
        else:
            out_writer.write(result)

    if out_writer is None:
        cv2.destroyAllWindows()
        cv2.waitKey(1)
    else:
        out_writer.release()


def pad_image_and_put_caption(
    image: np.ndarray,
    run_name: int,
    model_1: str,
    model_1_score: str,
    model_2: Optional[str] = None,
    model_2_score: Optional[str] = None,
    fps: Optional[int] = None,
) -> np.ndarray:
    """
    Pad the image with white and put the caption on it.

    :param image: Numpy array containing the image to be padded.
    :param run_name: Experiment description.
    :param model_1: Name of the model 1.
    :param model_1_score: Score of the model 1.
    :param model_2: Name of the model 2.
    :param model_2_score: Score of the model 2.
    :param fps: FPS of the inference.
    :return: Padded image with caption.
    """
    # Calculate text and image padding size
    text_scale = round(image.shape[1] / 1280, 1)
    thickness = int(text_scale / 1.5)
    (_, label_height), baseline = cv2.getTextSize(
        "Test caption", cv2.FONT_HERSHEY_SIMPLEX, text_scale, thickness
    )
    universal_padding = 2
    bottom_padding_pre_line = label_height + baseline
    # Prepare image captions
    caption_lines = [
        run_name + ("" if fps is None else f" @{fps} fps"),
        f"Model 1: {model_1}, score {model_1_score:.2f}",
    ]
    if model_2 and model_2_score:
        caption_lines.append(f"Model 2: {model_2}, score {model_2_score:.2f}")
    # Pad the image and put captions on it
    padded_image = cv2.copyMakeBorder(
        image,
        top=universal_padding,
        bottom=universal_padding + bottom_padding_pre_line * len(caption_lines),
        left=universal_padding,
        right=universal_padding,
        borderType=cv2.BORDER_CONSTANT,
        value=(255, 255, 255),
    )
    # Put text
    for line_number, text_line in enumerate(caption_lines):
        cv2.putText(
            padded_image,
            text_line,
            (0, image.shape[0] + bottom_padding_pre_line * (line_number + 1)),
            cv2.FONT_HERSHEY_SIMPLEX,
            text_scale,
            (0, 0, 0),
            thickness,
        )
    return padded_image


def concat_prediction_results(results: List[List[np.ndarray]]) -> np.ndarray:
    """
    Merge the prediction images to one.

    :param results: List of lists of numpy arrays containing the results of the
        predictions.
    :return: Numpy array containing the concatenated results.
    """
    # Gather information about images on the grid
    row_pixel_lengths = []
    for index, row in enumerate(results):
        integral_row_length = sum([image.shape[1] for image in row])
        row_pixel_lengths.append(integral_row_length)
        image_heights = [image.shape[0] for image in row]
        if len(set(image_heights)) > 1:
            raise ValueError(f"Row {index} has images with different heights!")
    # Concatenate images
    max_row_length = max(row_pixel_lengths)
    concatenated_rows = []
    for row in results:
        merged_row = cv2.hconcat(row)
        if merged_row.shape[1] < max_row_length:
            # Add empty image to the end of the row
            merged_row = cv2.hconcat(
                [
                    merged_row,
                    np.ones(
                        (
                            merged_row.shape[0],
                            max_row_length - merged_row.shape[1],
                            merged_row.shape[2],
                        ),
                        dtype=np.uint8,
                    )
                    * 255,
                ]
            )
        concatenated_rows.append(merged_row)
    return cv2.vconcat(concatenated_rows)



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import platform
import sys
from contextlib import contextmanager
from logging.handlers import MemoryHandler
from typing import Dict, List, Optional, Sequence, Union

import cv2
import numpy as np
from openvino.runtime import get_version

import geti_sdk
from geti_sdk.data_models import Image, Video
from geti_sdk.geti import DEFAULT_LOG_FORMAT
from geti_sdk.http_session import GetiSession

try:
    # Updated OpenVINO API, from v2023.2.0
    from openvino.core import Core
    from openvino.properties import device as ov_device
except ImportError:
    # The old API
    from openvino.runtime import Core
    from openvino.runtime.properties import device as ov_device


def get_system_info(device: str = "CPU") -> Dict[str, str]:
    """
    Retrieve relevant information about the system

    :param device: Hardware device to retrieve the info for, for example "CPU", "GPU",
        etc. Defaults to "CPU".
    :return: Dictionary containing system information
        - Operating system
        - Processor name
        - Python version
        - Geti SDK version
        - OpenVINO version
    """
    ov_core = Core()
    try:
        device_info = ov_core.get_property(
            device_name=device, property=ov_device.full_name()
        )
    except RuntimeError as e:
        logging.warning(
            f"Unable to retrieve device info for device `{device}`. Failed with "
            f"error: `{e}`"
        )
        device_info = device
    info: Dict[str, str] = {}
    info["operating_system"] = platform.system()
    info["device_info"] = device_info
    info["python"] = platform.python_version()
    info["geti-sdk"] = geti_sdk.__version__
    info["openvino"] = get_version()
    return info


def load_benchmark_media(
    session: GetiSession,
    images: Optional[Sequence[Union[Image, np.ndarray, os.PathLike]]] = None,
    video: Optional[Union[Video, os.PathLike]] = None,
    frames: int = 200,
) -> List[np.ndarray]:
    """
    Load and standardize a list of media. This method will return a list of numpy 2D
    arrays containing the frame data. The list will contain exactly `frames` elements.

    Either `video` or `images` should be specified, but not both.

    If the list of `images` contains less than `frames` elements, the first image in
    the list will be used to pad the list up to a length of `frames`.
    Similarly, if the `video` contains less frames than what is specified in `frames`,
    the first frame of the video will be used as padding.

    :param session: GetiSession pointing to the Geti server that can be used to
        retrieve the media data
    :param images: List of either:
        - Geti Image objects, identifying images on the Geti server
        - numpy 2D arrays, containing the image data
        - PathLike objects, pointing to the filepaths of images to load
    :param video: Either a Geti Video object or a filepath pointing to a video file on
        disk.
        NOTE: Either `video` or `images` should be specified, but not both
    :param frames: Number of frames to return
    :return: List of 2D arrays containing image/frame data. The returned list has a
        length of exactly `frames`.
    """
    if video is not None and images is not None:
        raise ValueError("Please specify either `video` or `images`, but not both.")
    if frames > 1000:
        logging.warning(
            f"You have specified {frames} to be loaded into memory. This may cause "
            f"out of memory problems on systems with constrained resources. Please "
            f"consider reducing the number of frames to load"
        )
    loaded_frames: List[np.ndarray] = []
    if images is not None:
        if len(images) < frames:
            # Fill the list of images up to `frames` using the first image
            images += [images[0]] * (frames - len(images))
        elif len(images) > frames:
            # Select only the first `frames` images
            images = images[0:frames]
        for image_object in images:
            if isinstance(image_object, Image):
                image_np = image_object.get_data(session=session)
                loaded_frames.append(image_np)
            elif isinstance(image_object, np.ndarray):
                loaded_frames.append(image_object)
            elif isinstance(image_object, (os.PathLike, str)):
                image_cv = cv2.imread(image_object)
                image_cv = cv2.cvtColor(image_cv, cv2.COLOR_BGR2RGB)
                loaded_frames.append(image_cv)
            else:
                raise TypeError(
                    f"Encountered invalid image object of type {type(image_object)}. "
                    f"Please pass either Geti Image objects, numpy arrays with image "
                    f"data or image filepaths."
                )
    elif video is not None:
        video_data_path: str
        if isinstance(video, Video):
            video_data_path = video.get_data(session=session)
        elif isinstance(video, os.PathLike):
            video_data_path = video
        else:
            raise TypeError(
                f"Encountered invalid video object of type {video.type}. Please "
                f"pass either a Geti Video object or a filepath to a video on disk."
            )
        cap = cv2.VideoCapture(video_data_path)

        if cap is None or not cap.isOpened():
            raise ValueError(f"Unable to read video from {video_data_path}")

        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        while cap.isOpened() and len(loaded_frames) < frame_count:
            ret, frame = cap.read()
            if ret is True:
                rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                loaded_frames.append(rgb_frame)
            else:
                break
        cap.release()
        if len(loaded_frames) < frame_count:
            # Pad with first frame if the video was too short
            loaded_frames += [loaded_frames[0]] * (frames - len(loaded_frames))
    return loaded_frames


@contextmanager
def suppress_log_output(
    target_logger: Optional[logging.Logger] = None,
    target_handler: Optional[logging.Handler] = None,
):
    """
    Context manager to temporarily capture and suppress log output.
    All logging calls made in the context will be redirected to memory,
    unless ERROR level message are logged. In that case the output will be sent to
    the `target_handler`.

    :param target_logger: Optional Logger of which the output should be captured
    :param target_handler: Optional Handler that should be used in case of ERROR log
        message
    """
    if target_logger is None:
        target_logger = logging.getLogger()

    if target_handler is None:
        target_handler = logging.StreamHandler(stream=sys.stdout)
    formatter = logging.Formatter(DEFAULT_LOG_FORMAT)
    target_handler.setFormatter(formatter)

    memory_handler = MemoryHandler(
        capacity=1000, flushLevel=logging.ERROR, target=target_handler
    )

    original_handlers = target_logger.handlers
    for handler in original_handlers:
        target_logger.removeHandler(handler)
    target_logger.addHandler(memory_handler)

    try:
        yield target_logger
    except Exception:
        target_logger.exception("Error in `suppress_log_output`")
        raise
    finally:
        # Flush the memory handler and reinstate the original log handlers
        super(MemoryHandler, memory_handler).flush()
        target_logger.removeHandler(memory_handler)
        for handler in original_handlers:
            target_logger.addHandler(handler)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from importlib import resources

try:
    DEFAULT_DATA_PATH = str(resources.files("geti_sdk.demos") / "data")
except AttributeError:
    with resources.path("geti_sdk.demos", "data") as data_path:
        DEFAULT_DATA_PATH = str(data_path)

EXAMPLE_IMAGE_PATH = os.path.join(DEFAULT_DATA_PATH, "example", "dogs.png")



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import datetime
import json
import logging
import os
import shutil
from collections.abc import Sequence
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import attr
import numpy as np
from model_api.models import Model as OMZModel

from geti_sdk.data_models import Label, Prediction, Project, Task
from geti_sdk.deployment.data_models import ROI, IntermediateInferenceResult
from geti_sdk.rest_converters import ProjectRESTConverter

from .deployed_model import DeployedModel
from .inference_hook_interfaces import PostInferenceHookInterface
from .utils import (
    OVMS_README_PATH,
    assign_empty_label,
    flow_control,
    generate_ovms_model_name,
    pipeline_callback_factory,
)


@attr.define(slots=False)
class Deployment:
    """
    Representation of a deployed Intel® Geti™ project that can be used to run
    inference locally
    """

    project: Project
    models: List[DeployedModel]

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._is_single_task: bool = len(self.project.get_trainable_tasks()) == 1
        self._are_models_loaded: bool = False
        self._inference_converters: Dict[str, Any] = {}
        self._path_to_temp_resources: Optional[str] = None
        self._requires_resource_cleanup: bool = False
        self._post_inference_hooks: List[PostInferenceHookInterface] = []
        self._empty_label: Optional[Label] = None
        self._asynchronous_mode: bool = False
        self._async_callback_defined: bool = False

    @property
    def is_single_task(self) -> bool:
        """
        Return True if the deployment represents a project with only a single task.

        :return: True if the deployed project contains only one trainable task, False
            if it is a pipeline project
        """
        return self._is_single_task

    @property
    def are_models_loaded(self) -> bool:
        """
        Return True if all inference models for the Deployment are loaded and ready
        to infer.

        :return: True if all inference models for the deployed project are loaded in
            memory and ready for inference
        """
        return self._are_models_loaded

    @property
    def asynchronous_mode(self) -> bool:
        """
        Return True if the deployment is configured for asynchronous inference execution.

        Asynchronous execution can result in a large increase in throughput for
        certain applications, for example video processing. However, it requires
        slightly more configuration compared to synchronous (the default) mode. For a
        more detailed overview of the differences between synchronous and
        asynchronous execution, please refer to the OpenVINO documentation at
        https://docs.openvino.ai/2024/notebooks/115-async-api-with-output.html

        :return: True if the deployment is set in asynchronous execution mode, False
            if it is in synchronous mode.
        """
        return self._async_callback_defined

    @asynchronous_mode.setter
    def asynchronous_mode(self, mode: bool):
        """
        Set the inference mode for the deployment

        :param mode: False to set the deployment to synchronous mode. Removes all
            asynchronous callbacks for the models in the deployment.
        """
        if mode:
            if not self._async_callback_defined:
                raise ValueError(
                    "Please use the `set_asynchronous_callback` method to switch a "
                    "deployment to asynchronous inference mode."
                )
            else:
                logging.debug("Deployment is already in asynchronous mode.")
        else:
            self._async_callback_defined = False
            for model in self.models:
                model.asynchronous_mode = False
            logging.info("Asynchronous inference mode disabled. All callbacks removed.")

    def save(self, path_to_folder: Union[str, os.PathLike]) -> bool:
        """
        Save the Deployment instance to a folder on local disk.

        :param path_to_folder: Folder to save the deployment to
        :return: True if the deployment was saved successfully, False otherwise
        """
        project_dict = ProjectRESTConverter.to_dict(self.project, deidentify=False)
        deployment_folder = os.path.join(path_to_folder, "deployment")

        os.makedirs(deployment_folder, exist_ok=True, mode=0o770)

        # Save model for each task
        for task, model in zip(self.project.get_trainable_tasks(), self.models):
            model_dir = os.path.join(deployment_folder, task.title)
            os.makedirs(model_dir, exist_ok=True, mode=0o770)
            success = model.save(model_dir)
            if not success:
                logging.exception(
                    f"Saving model '{model.name}' failed. Unable to save deployment."
                )
                return False

        # Save project data
        project_filepath = os.path.join(deployment_folder, "project.json")
        with open(project_filepath, "w") as project_file:
            json.dump(project_dict, project_file, indent=4)

        # Save post inference hooks, if any
        if self.post_inference_hooks:
            hook_config_file = os.path.join(deployment_folder, "hook_config.json")
            hook_configs: List[Dict[str, Any]] = []
            for hook in self.post_inference_hooks:
                hook_configs.append(hook.to_dict())
            with open(hook_config_file, "w") as file:
                json.dump({"post_inference_hooks": hook_configs}, file)

        # Clean up temp resources if needed
        if self._requires_resource_cleanup:
            self._remove_temporary_resources()
            self._requires_resource_cleanup = False

        return True

    @classmethod
    def from_folder(cls, path_to_folder: Union[str, os.PathLike]) -> "Deployment":
        """
        Create a Deployment instance from a specified `path_to_folder`.

        :param path_to_folder: Path to the folder containing the Deployment data
        :return: Deployment instance corresponding to the deployment data in the folder
        """
        deployment_folder = path_to_folder
        if not isinstance(path_to_folder, str):
            path_to_folder = str(path_to_folder)
        if not path_to_folder.endswith("deployment"):
            if "deployment" in os.listdir(path_to_folder):
                deployment_folder = os.path.join(path_to_folder, "deployment")
            else:
                raise ValueError(
                    f"No `deployment` folder found in the directory at "
                    f"`{path_to_folder}`. Unable to load Deployment."
                )
        project_filepath = os.path.join(deployment_folder, "project.json")
        with open(project_filepath, "r") as project_file:
            project_dict = json.load(project_file)
        project = ProjectRESTConverter.from_dict(project_dict)
        task_folder_names = [task.title for task in project.get_trainable_tasks()]
        models: List[DeployedModel] = []
        for task_folder in task_folder_names:
            models.append(
                DeployedModel.from_folder(os.path.join(deployment_folder, task_folder))
            )
        deployment = cls(models=models, project=project)

        # Load post inference hooks, if any
        hook_config_file = os.path.join(deployment_folder, "hook_config.json")
        if os.path.isfile(hook_config_file):
            available_hooks = {
                subcls.__name__: subcls
                for subcls in PostInferenceHookInterface.__subclasses__()
            }
            with open(hook_config_file, "r") as file:
                hook_dict = json.load(file)
            for hook_data in hook_dict["post_inference_hooks"]:
                for hook_name, hook_args in hook_data.items():
                    target_hook = available_hooks[hook_name]
                    hook = target_hook.from_dict(hook_args)
                deployment.add_post_inference_hook(hook)
        return deployment

    def load_inference_models(
        self,
        device: Union[str, Sequence[str]] = "CPU",
        max_async_infer_requests: Optional[Union[int, Sequence[int]]] = None,
        openvino_configuration: Optional[Dict[str, str]] = None,
    ):
        """
        Load the inference models for the deployment to the specified device.

        Note: For a list of devices that are supported for OpenVINO inference, please see:
        https://docs.openvino.ai/latest/openvino_docs_OV_UG_supported_plugins_Supported_Devices.html

        :param device: Device to load the inference models to (e.g. 'CPU', 'GPU',
            'AUTO', etc).

            **NOTE**: For task chain deployments, it is possible to pass a list of device names
            instead. Each entry in the list is the target device for the model
            corresponding to it's index. I.e. the first entry is applied for the
            first model, the second entry for the second model.
        :param max_async_infer_requests: Maximum number of infer requests to use in
            asynchronous mode. This parameter only takes effect when the asynchronous
            inference mode is used. It controls the maximum number of request that
            will be handled in parallel. When set to 0, OpenVINO will attempt to
            determine the optimal number of requests for your system automatically.
            When left as None (the default), a single infer request per model will be
            used to conserve memory

            **NOTE**: For task chain deployments, it is possible to pass a list of integers.
            Each entry in the list is the maximum number of infer requests for the model
            corresponding to it's index. I.e. the first number is applied for the
            first model, the second number for the second model.
        :param openvino_configuration: Configuration for the OpenVINO execution mode
            and plugins. This can include for example specific performance hints. For
            further details, refer to the OpenVINO documentation here:
            https://docs.openvino.ai/2022.3/openvino_docs_OV_UG_Performance_Hints.html#doxid-openvino-docs-o-v-u-g-performance-hints
        """
        if max_async_infer_requests is None:
            max_async_infer_requests = 1
        elif max_async_infer_requests == 0 and device == "CPU":
            max_async_infer_requests = os.cpu_count()

        for idx, model in enumerate(self.models):
            if isinstance(device, Sequence) and not isinstance(device, str):
                dev = device[idx]
            else:
                dev = device
            if isinstance(max_async_infer_requests, Sequence):
                max_requests = max_async_infer_requests[idx]
            else:
                max_requests = max_async_infer_requests
            model.load_inference_model(
                device=dev,
                project=self.project,
                max_async_infer_requests=max_requests,
                plugin_configuration=openvino_configuration,
                task_index=idx,
            )

        # Extract empty label for the upstream task
        self._empty_label = self.models[0].labels.get_empty_label()

        self._are_models_loaded = True
        logging.info(f"Inference models loaded on device `{device}` successfully.")

    def infer(self, image: np.ndarray, name: Optional[str] = None) -> Prediction:
        """
        Run inference on an image for the full model chain in the deployment.

        :param image: Image to run inference on, as a numpy array containing the pixel
            data. The image is expected to have dimensions [height x width x channels],
            with the channels in RGB order
        :param name: Optional name for the image, if specified this will be used in
            any post inference hooks belonging to the deployment.
        :return: inference results
        """
        self._check_models_loaded()

        if self.asynchronous_mode:
            raise RuntimeError(
                "Unable to use synchronous inference while the deployment is set to "
                "asynchronous execution mode. Please use the `.infer_async` method "
                "instead."
            )

        # Single task inference
        if self.is_single_task:
            prediction = self._infer_task(
                image, task=self.project.get_trainable_tasks()[0], explain=False
            )
        # Multi-task inference
        else:
            prediction = self._infer_pipeline(image=image, explain=False)

        # Empty label is not generated by prediction postprocessing correctly, append
        # it here if there are no other predictions
        assign_empty_label(
            prediction=prediction, image=image, empty_label=self._empty_label
        )

        self._execute_post_inference_hooks(
            image=image, prediction=prediction, name=name
        )

        return prediction

    def infer_async(
        self, image: np.ndarray, runtime_data: Optional[Any] = None
    ) -> None:
        """
        Perform asynchronous inference on the `image`.

        **NOTE**: Inference results are not returned directly! Instead, a
        post-inference callback should be defined to handle results, using the
        `.set_asynchronous_callback` method.

        :param image: numpy array representing an image
        :param runtime_data: An optional object containing any additional data
            that should be passed to the asynchronous callback for each infer request.
            This can for example be a timestamp or filename for the image to infer.
            Passing complex objects like a tuple/list or dictionary is also supported.
        """
        self._check_models_loaded()

        if not self.asynchronous_mode:
            raise RuntimeError(
                "No callback function defined to handle asynchronous inference, "
                "please make sure to define a callback using "
                "the `.set_asynchronous_callback()` method of the deployment, "
                "otherwise your inference results will be lost."
            )

        runtime_data_tuple = (image, runtime_data)

        # Note that this method is the same for both single task and pipeline projects
        # Infer requests for downstream models in a pipeline are handled via callbacks
        self._infer_task_async(
            image,
            task=self.project.get_trainable_tasks()[0],
            explain=False,
            runtime_data=runtime_data_tuple,
        )

    def explain(self, image: np.ndarray, name: Optional[str] = None) -> Prediction:
        """
        Run inference on an image for the full model chain in the deployment. The
        resulting prediction will also contain saliency maps and the feature vector
        for the input image.

        :param image: Image to run inference on, as a numpy array containing the pixel
            data. The image is expected to have dimensions [height x width x channels],
            with the channels in RGB order
        :param name: Optional name for the image, if specified this will be used in
            any post inference hooks belonging to the deployment.
        :return: inference results
        """
        self._check_models_loaded()

        if self.asynchronous_mode:
            raise RuntimeError(
                "Unable to use synchronous inference while the deployment is set to "
                "asynchronous execution mode. Please use the `.explain_async` method "
                "instead."
            )

        # Single task inference
        if self.is_single_task:
            prediction = self._infer_task(
                image, task=self.project.get_trainable_tasks()[0], explain=True
            )
        # Multi-task inference
        else:
            prediction = self._infer_pipeline(image=image, explain=True)

        self._execute_post_inference_hooks(
            image=image, prediction=prediction, name=name
        )
        return prediction

    def explain_async(
        self, image: np.ndarray, runtime_data: Optional[Any] = None
    ) -> None:
        """
        Perform asynchronous inference on the `image`, and generate saliency maps and
        feature vectors

        **NOTE**: Inference results are not returned directly! Instead, a
        post-inference callback should be defined to handle results, using the
        `.set_asynchronous_callback` method.

        :param image: numpy array representing an image
        :param runtime_data: An optional object containing any additional data
            that should be passed to the asynchronous callback for each infer request.
            This can for example be a timestamp or filename for the image to infer.
            Passing complex objects like a tuple/list or dictionary is also supported.
        """
        self._check_models_loaded()

        if not self.asynchronous_mode:
            raise RuntimeError(
                "No callback function defined to handle asynchronous inference, "
                "please make sure to define a callback using "
                "the `.set_asynchronous_callback()` method of the deployment, "
                "otherwise your inference results will be lost."
            )

        runtime_data_tuple = (image, runtime_data)

        # Note that this method is the same for both single task and pipeline projects
        # Infer requests for downstream models in a pipeline are handled via callbacks
        self._infer_task_async(
            image,
            task=self.project.get_trainable_tasks()[0],
            explain=True,
            runtime_data=runtime_data_tuple,
        )

    def _check_models_loaded(self) -> None:
        """
        Check if models are loaded and ready for inference.

        :raises: ValueError in case models are not loaded
        """
        if not self.are_models_loaded:
            raise ValueError(
                f"Deployment '{self}' is not ready to infer, the inference models are "
                f"not loaded. Please call 'load_inference_models' first."
            )

    def _infer_task(
        self, image: np.ndarray, task: Task, explain: bool = False
    ) -> Prediction:
        """
        Run pre-processing, inference, and post-processing on the input `image`, for
        the model associated with the `task`.

        :param image: Image to run inference on
        :param task: Task to run inference for
        :param explain: True to get additional outputs for model explainability,
            including saliency maps and the feature vector for the image
        :return: Inference result
        """
        model = self._get_model_for_task(task)
        return model.infer(image, explain)

    def _infer_task_async(
        self,
        image: np.ndarray,
        task: Task,
        explain: bool = False,
        runtime_data: Optional[Any] = None,
    ) -> None:
        """
        Perform asynchronous inference on the `image`.

        **NOTE**: Inference results are not returned directly! Instead, a
        post-inference callback should be defined to handle results, using the
        `.set_asynchronous_callback` method.

        :param image: numpy array representing an image
        :param task: Task to run inference for
        :param explain: True to include saliency maps and feature maps in the returned
            Prediction. Note that these are only available if supported by the model.
        :param runtime_data: An optional object containing any additional data
            that should be passed to the asynchronous callback for each infer request.
            This can for example be a timestamp or filename for the image to infer.
            Passing complex objects like a tuple/list or dictionary is also supported.
        """
        model = self._get_model_for_task(task)
        model.infer_async(image=image, explain=explain, runtime_data=runtime_data)

    def _infer_pipeline(self, image: np.ndarray, explain: bool = False) -> Prediction:
        """
        Run pre-processing, inference, and post-processing on the input `image`, for
        all models in the task chain associated with the deployment.

        Note: If `explain=True`, a saliency map, feature vector and active score for
        the first task in the pipeline will be included in the prediction output

        :param image: Image to run inference on
        :param explain: True to get additional outputs for model explainability,
            including saliency maps and the feature vector for the image
        :return: Inference result
        """
        previous_labels: Optional[List[Label]] = None
        intermediate_result: Optional[IntermediateInferenceResult] = None
        rois: Optional[List[ROI]] = None
        image_views: Optional[List[np.ndarray]] = None

        # Pipeline inference
        for task in self.project.pipeline.tasks[1:]:
            # First task in the pipeline generates the initial result and ROIs
            if task.is_trainable and previous_labels is None:
                task_prediction = self._infer_task(image, task=task, explain=explain)
                rois: Optional[List[ROI]] = None
                if not task.is_global:
                    rois = [
                        ROI.from_annotation(annotation)
                        for annotation in task_prediction.annotations
                    ]
                intermediate_result = IntermediateInferenceResult(
                    image=image, prediction=task_prediction, rois=rois
                )
                previous_labels = [label for label in task.labels if not label.is_empty]

            # Downstream trainable tasks
            elif task.is_trainable:
                if rois is None or image_views is None or intermediate_result is None:
                    raise NotImplementedError(
                        "Unable to run inference for the pipeline in the deployed "
                        "project: A flow control task is required between each "
                        "trainable task in the pipeline."
                    )
                new_rois: List[ROI] = []
                for roi, view in zip(rois, image_views):
                    view_prediction = self._infer_task(view, task)
                    if task.is_global:
                        # Global tasks add their labels to the existing shape in the ROI
                        intermediate_result.extend_annotations(
                            view_prediction.annotations, roi=roi
                        )
                    else:
                        # Local tasks create new shapes in the image coordinate system,
                        # and generate ROI's corresponding to the new shapes
                        for annotation in view_prediction.annotations:
                            intermediate_result.append_annotation(annotation, roi=roi)
                            new_rois.append(ROI.from_annotation(annotation))
                        intermediate_result.rois = [
                            new_roi.to_absolute_coordinates(parent_roi=roi)
                            for new_roi in new_rois
                        ]
                previous_labels = [label for label in task.labels if not label.is_empty]

            # Downstream flow control tasks
            else:
                if previous_labels is None:
                    raise NotImplementedError(
                        f"Unable to run inference for the pipeline in the deployed "
                        f"project: First task in the pipeline after the DATASET task "
                        f"has to be a trainable task, found task of type {task.type} "
                        f"instead."
                    )
                rois, image_views = flow_control(
                    intermediate_result=intermediate_result, task=task
                )
        return intermediate_result.prediction

    def _get_model_for_task(self, task: Task) -> DeployedModel:
        """
        Get the DeployedModel instance corresponding to the input `task`.

        :param task: Task to get the model for
        :return: DeployedModel corresponding to the task
        """
        try:
            task_index = self.project.get_trainable_tasks().index(task)
        except ValueError as error:
            raise ValueError(
                f"Task {task.title} is not in the list of trainable tasks for project "
                f"{self.project.name}."
            ) from error
        return self.models[task_index]

    def _remove_temporary_resources(self) -> bool:
        """
        If necessary, clean up any temporary resources associated with the deployment.

        :return: True if temp files have been deleted successfully
        """
        if self._path_to_temp_resources is not None and os.path.isdir(
            self._path_to_temp_resources
        ):
            try:
                shutil.rmtree(self._path_to_temp_resources)
            except PermissionError:
                logging.warning(
                    f"Unable to remove temporary files for deployment at path "
                    f"`{self._path_to_temp_resources}` because the files are in "
                    f"use by another process. "
                )
                return False
        else:
            logging.debug(
                f"Unable to clean up temporary resources for deployment {self}, "
                f"because the resources were not found on the system. Possibly "
                f"they were already deleted."
            )
            return False
        return True

    def __del__(self):
        """
        If necessary, clean up any temporary resources associated with the deployment.
        This method is called when the Deployment instance is deleted.
        """
        if self._requires_resource_cleanup:
            self._remove_temporary_resources()

    def generate_ovms_config(self, output_folder: Union[str, os.PathLike]) -> None:
        """
        Generate the configuration files needed to push the models for the
        `Deployment` instance to OVMS.

        :param output_folder: Target folder to save the configuration files to
        """
        # First prepare the model config list
        if os.path.basename(output_folder) != "ovms_models":
            ovms_models_dir = os.path.join(output_folder, "ovms_models")
        else:
            ovms_models_dir = output_folder
            output_folder = os.path.dirname(ovms_models_dir)
        os.makedirs(ovms_models_dir, exist_ok=True)

        model_configs: List[Dict[str, Dict[str, Any]]] = []
        for model in self.models:
            # Create configuration entry for model
            model_name = generate_ovms_model_name(
                project=self.project, model=model, omit_version=True
            )
            config = {
                "name": model_name,
                "base_path": f"/models/{model_name}",
                "plugin_config": {"PERFORMANCE_HINT": "LATENCY"},
            }
            model_configs.append({"config": config})

            # Copy IR model files to the expected OVMS format
            if model.version is not None:
                model_version = str(model.version)
            else:
                # Fallback to version 1 if no version info is available
                model_version = "1"

            ovms_model_dir = os.path.join(ovms_models_dir, model_name, model_version)

            # Load the model to embed preprocessing for inference with OVMS adapter
            embedded_model = OMZModel.create_model(
                model=os.path.join(model.model_data_path, "model.xml")
            )
            embedded_model.save(
                xml_path=os.path.join(ovms_model_dir, "model.xml"),
                bin_path=os.path.join(ovms_model_dir, "model.bin"),
            )
            logging.info(f"Model `{model.name}` prepared for OVMS inference.")

        # Save model configurations
        ovms_config_list = {"model_config_list": model_configs}
        config_target_filepath = os.path.join(ovms_models_dir, "ovms_model_config.json")
        with open(config_target_filepath, "w") as file:
            json.dump(ovms_config_list, file)

        # Copy resource files
        shutil.copy2(OVMS_README_PATH, os.path.join(output_folder))

        logging.info(
            f"Configuration files for OVMS model deployment have been generated in "
            f"directory '{output_folder}'. This folder contains a `OVMS_README.md` "
            f"file with instructions on how to launch OVMS, connect to it and run "
            f"inference. Please follow the instructions outlined there to get started."
        )

    def infer_queue_full(self) -> bool:
        """
        Return True if the queue for asynchronous infer requests is full, False
        otherwise

        :return: True if the infer queue is full, False otherwise
        """
        if not self.asynchronous_mode:
            logging.warning(
                "Method `infer_queue_full()` has no effect in synchronous execution "
                "mode"
            )
        # Check the infer queue of the first model in the deployment
        return self.models[0].infer_queue_full()

    def await_all(self) -> None:
        """
        Block execution until all asynchronous infer requests have finished
        processing.

        This means that program execution will resume once the infer queue is empty

        This is a flow control function, it is only applicable when using
        asynchronous inference.
        """
        if not self.asynchronous_mode:
            logging.warning(
                "Method `await_all()` has no effect in synchronous execution " "mode"
            )
        # Wait until the infer queue of the last model in the task chain is empty
        for model in self.models:
            model.await_all()

    def await_any(self) -> None:
        """
        Block execution until any of the asynchronous infer requests currently in
        the infer queue completes processing

        This means that program execution will resume once a single spot becomes
        available in the infer queue

        This is a flow control function, it is only applicable when using
        asynchronous inference.
        """
        if not self.asynchronous_mode:
            logging.warning(
                "Method `await_any()` has no effect in synchronous execution " "mode"
            )
        self.models[0].await_any()

    def set_asynchronous_callback(
        self,
        callback_function: Optional[
            Callable[[np.ndarray, Prediction, Optional[Any]], None]
        ] = None,
    ) -> None:
        """
        Set the callback function to handle asynchronous inference results. This
        function is called whenever a result for an asynchronous inference request
        comes available.

        **NOTE**: Calling this method enables asynchronous inference mode for the
                  deployment. The regular synchronous inference method will no longer
                  be available, unless the deployment is reloaded.

        :param callback_function: Function that should be called to handle
            asynchronous inference results. The function should take the following
            input parameters:

             1. The image/video frame. This is the original image to infer
             2. The inference results (the Prediction). This is the model output for
                the image
             2. Any additional data that will be passed with the infer
                request at runtime. For example, this could be a timestamp for the
                frame, or a title/filepath, etc. This can be in the form of any object:
                You can for instance pass a dictionary, or a tuple/list of multiple
                objects

            **NOTE**: It is possible to call this method without specifying any
                      callback function. In that case, the deployment will be switched
                      to asynchronous mode but only the post-inference hooks will be
                      executed after each infer request
        """

        def post_inference_hook_callback(
            image: np.ndarray, prediction: Prediction, runtime_data: Optional[Any]
        ):
            if isinstance(runtime_data, dict):
                name = runtime_data.get("name", None)
            else:
                name = None
            self._execute_post_inference_hooks(
                image=image, prediction=prediction, name=name
            )

        if callback_function is None:
            final_callback_function = post_inference_hook_callback
        else:

            def callback_and_hook(image, prediction, runtime_data):
                post_inference_hook_callback(image, prediction, runtime_data)
                callback_function(image, prediction, runtime_data)

            final_callback_function = callback_and_hook

        if self.is_single_task:

            def full_callback(result: Prediction, runtime_data: Tuple[np.ndarray, Any]):
                image, runtime_user_data = runtime_data
                assign_empty_label(
                    prediction=result, image=image, empty_label=self._empty_label
                )

                # User defined callback to further process the prediction results
                final_callback_function(image, result, runtime_user_data)

            self.models[0].set_asynchronous_callback(full_callback)
        else:
            # Pipeline case, this is a general implementation accounting for
            # pipelines with more than 2 trainable tasks!
            callbacks: List[Callable[[Prediction, Any], None]] = []
            trainable_tasks = self.project.get_trainable_tasks()

            for task_idx, task in enumerate(self.project.pipeline.tasks):
                if not task.is_trainable:
                    # No callbacks are added for flow control tasks
                    continue

                trainable_task_idx = trainable_tasks.index(task)
                is_first_task = trainable_task_idx == 0
                is_final_task = trainable_task_idx == len(trainable_tasks) - 1

                next_task: Optional[Task] = None
                next_trainable_task: Optional[Task] = None
                add_flow_control = False
                if not is_final_task:
                    next_task = self.project.pipeline.tasks[task_idx + 1]
                    next_trainable_task = trainable_tasks[trainable_task_idx + 1]
                    add_flow_control = next_task.type.is_trainable

                callbacks.append(
                    pipeline_callback_factory(
                        is_first_task=is_first_task,
                        is_final_task=is_final_task,
                        add_flow_control=add_flow_control,
                        current_task=task,
                        deployment=self,
                        next_trainable_task=next_trainable_task,
                        next_task=next_task,
                        final_callback=final_callback_function,
                    )
                )
            for task, callback in zip(trainable_tasks, callbacks):
                model = self._get_model_for_task(task)
                model.set_asynchronous_callback(callback)

        self._async_callback_defined = True
        logging.info("Asynchronous inference mode enabled.")

    @property
    def post_inference_hooks(self) -> List[PostInferenceHookInterface]:
        """
        Return the currently active post inference hooks for the deployment

        :return: list of PostInferenceHook objects
        """
        return self._post_inference_hooks

    def clear_inference_hooks(self) -> None:
        """
        Remove all post inference hooks for the deployment
        """
        n_hooks = len(self.post_inference_hooks)
        self._post_inference_hooks = []
        if n_hooks != 0:
            logging.info(
                f"Post inference hooks cleared. {n_hooks} hooks were removed "
                f"successfully"
            )

    def add_post_inference_hook(self, hook: PostInferenceHookInterface) -> None:
        """
        Add a post inference hook, which will be executed after each call to
        `Deployment.infer`

        :param hook: PostInferenceHook to be added to the deployment
        """
        self._post_inference_hooks.append(hook)
        logging.info(f"Hook `{hook}` added.")
        logging.info(
            f"Deployment now contains {len(self.post_inference_hooks)} "
            f"post inference hooks."
        )

    def _execute_post_inference_hooks(
        self, image: np.ndarray, prediction: Prediction, name: Optional[str] = None
    ) -> None:
        """
        Execute all post inference hooks

        :param image: Numpy image which was inferred
        :param prediction: Prediction for the image
        :param name: Optional name for the image
        """
        timestamp = datetime.datetime.now()
        for hook in self._post_inference_hooks:
            hook.run(image, prediction, name, timestamp)
